"""
session
~~~~~~~~~~
Service Session, low-level object for communicating with RESTFul services.
"""

import requests
import requests.utils
from requests import codes
from os import environ
import copy
import json
import platform
import time
import logging
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry  # pylint: disable=E0401
from hashlib import sha1
from typing import Dict

from . import __version__
from .exceptions import ServerError, ServiceNotFound
from .utils import check_resp_error, get_access_token
from .config import Config, load_cache, save_cache

log = logging.getLogger(__name__)

config = Config()


class ServiceSession(requests.Session):
    """
    A wrapped requests session object with base_url and exported endpoints
    from nextcode service api's
    """

    def __init__(self, url_base, api_key, *args, **kwargs):
        super(ServiceSession, self).__init__(*args, **kwargs)
        # retry idempotent methods up to 5 times
        if not environ.get("NEXTCODE_DISABLE_RETRY"):
            retries = 5
            backoff_factor = 0.5
            status_forcelist = (500, 502, 503, 504)
            retry = Retry(
                total=retries,
                read=retries,
                connect=retries,
                backoff_factor=backoff_factor,
                status_forcelist=status_forcelist,
            )
            adapter = HTTPAdapter(max_retries=retry)
            self.mount("http://", adapter)
            self.mount("https://", adapter)

        self.root_info = {}
        self.endpoints = {}
        self.token = None
        self.url_base = url_base
        self.api_key = api_key

        if environ.get("GOR_API_KEY"):
            log.info("Overriding api key from environment")
            self.api_key = environ["GOR_API_KEY"]

        self.cache_name = sha1((self.url_base).encode()).hexdigest()
        self.user_agent = "Nextcode-SDK/%s Python/%s %s/%s" % (
            __version__,
            platform.python_version(),
            platform.system(),
            platform.release(),
        )

        self.headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "User-Agent": self.user_agent,
        }
        if environ.get("NEXTCODE_ENABLE_GZIP"):
            self.headers["Accept-Encoding"] = "gzip"

        if not self._load():
            self._initialize()

        self.endpoints = self.root_info.get("endpoints")

    def _initialize(self) -> None:
        if environ.get('NEXTCODE_ACCESS_TOKEN'):
            self.token = environ.get('NEXTCODE_ACCESS_TOKEN')
        elif self.api_key:
            self.token = get_access_token(self.api_key)
        self.headers["Authorization"] = "Bearer {}".format(self.token)
        try:
            self.fetch_root_info()
        except ServerError as ex:
            if ex.response and ex.response.get("status") == codes.not_found:
                status = ex.response.get("status")
                raise ServiceNotFound(
                    f"Service does not exist on server ({status}): {self.url_base}"
                )
            raise
        # persist the endpoints to disk to save on a roundtrip every call
        self._save()

    def _save(self) -> None:
        contents = {
            "token": self.token,
            "root_info": self.root_info,
            "api_key": self.api_key,
        }
        save_cache(self.cache_name, contents)

    def _load(self) -> bool:
        contents = load_cache(self.cache_name)
        if not contents:
            return False
        self.token = contents["token"]
        self.root_info = contents["root_info"]
        if self.token:
            self.headers["Authorization"] = "Bearer {}".format(self.token)
        # if the service does not have our user available make sure to refresh
        if not self.root_info.get("current_user"):
            return False
        return True

    def _do_request(self, method, retry=True, *args, **kwargs):

        # method: GET
        old_content_type = self.headers["Content-Type"]
        
        if method == "get":
            # ! Temporary hack: Remove the application/json content-type header for GET's
            del self.headers["Content-Type"]

        try:
            st = time.time()
            response = getattr(super(ServiceSession, self), method)(*args, **kwargs)
            diff = time.time() - st
        finally:
            self.headers["Content-Type"] = old_content_type

        # Manage response from the server
        log.info(
            "%s %s returned %s in %.3f sec"
            % (method.upper(), args[0], response.status_code, diff)
        )
        if response.status_code == codes.unauthorized:
            if retry:
                log.debug("Status code %s, retrying once..." % response.status_code)
                self._initialize()
                log.info(
                    "Tokens have been updated. Now calling _do_request one more..."
                )
                return self._do_request(method, False, *args, **kwargs)
            else:
                log.error("Received unauthorized in retry: %s", response.text)

        check_resp_error(response)
        return response

    def get(self, *args, **kw):
        return self._do_request("get", True, *args, **kw)

    def put(self, *args, **kw):
        return self._do_request("put", True, *args, **kw)

    def post(self, *args, **kw):
        return self._do_request("post", True, *args, **kw)

    def delete(self, *args, **kw):
        return self._do_request("delete", True, *args, **kw)

    def fetch_root_info(self) -> Dict:
        log.debug(
            "fetch_root_info(): url_base: {0}, headers: {1}".format(
                self.url_base, self.headers
            )
        )
        try:
            r = requests.get(self.url_base, timeout=3.0, headers=self.headers)
        except requests.exceptions.ConnectionError as ex:
            raise ServerError(
                f"Could not reach server {self.url_base} ({ex})"
            ) from None
        except requests.exceptions.ReadTimeout as ex:
            raise ServerError(
                f"Could not reach server {self.url_base} ({ex})"
            ) from None

        r.raise_for_status()

        if "application/json" not in r.headers["Content-Type"]:
            raise ServerError(
                "Unexpected response: %s" % r.text, url=self.url_base
            ) from None
        ret = r.json()
        self.root_info = ret
        return ret

    def url_from_endpoint(self, endpoint: str) -> str:
        try:
            return self.endpoints[endpoint]
        except KeyError:
            raise ServerError(
                "Endpoint '%s' is not exported by '%s'.\nAvailable endpoints are %s"
                % (endpoint, self.url_base, ", ".join(self.endpoints.keys()))
            )

    def request(self, method, url, **kwargs):
        log.debug("Calling %s %s", method, url)
        stripped_headers = copy.copy(self.headers)
        stripped_headers["Authorization"] = "Bearer ***"
        log.debug("Headers: %s" % json.dumps(stripped_headers))
        if "json" in kwargs:
            log.debug("Payload:\n%s" % json.dumps(kwargs["json"], indent=4))
        return super(ServiceSession, self).request(method, url, **kwargs)

    def links(self, resp: Dict) -> Dict:
        return resp.get("links", {})



"""
Analysis catalog run
------------------

Analysis catalog run status


"""

import json
import datetime
import dateutil
import time
import logging
from posixpath import join as urljoin
from typing import Callable, Union, Optional, Dict, List

from .exceptions import PhenotypeError
from ...exceptions import ServerError
from ...session import ServiceSession

log = logging.getLogger(__name__)


class AnalysisCatalogRun:
    """
    A local object representing an analysis catalog run response from the phenotype
    catalog service.

    Please refer to the API Documentation for the phenotype catalog service.

    In addition to the ones documented here, this object has at least these attributes:

    * name - Analysis Catalog Run name
    * state - the state of the Analysis Catalog Run
    * created_at - Timestamp when the analysis catalog run was first created
    * updated_at - Timestamp when the analysis catalog run was last updated
    * ended_at - Timestamp when the analysis catalog run ended (entered 'failed' or 'completed' state)

    """

    def __init__(self, session: ServiceSession, data: Dict):
        self.session = session
        self.data = data

    def __getattr__(self, name):
        try:
            val = self.data[name]
        except KeyError:
            raise AttributeError

        try:
            val = dateutil.parser.parse(val)
        except Exception:
            pass

        return val

    def __repr__(self) -> str:
        return f"<AnalysisCatalogRun {self.name} in project {self.project_key}>"



"""
Exceptions
~~~~~~~~~~

Custom exceptions raised by the query service.

"""

from typing import List


class TemplateError(Exception):
    pass


class QueryError(Exception):
    def __init__(self, message: str, query_id: int = None):
        self.query_id = query_id


class MissingRelations(Exception):
    def __init__(self, relations: List):
        self.relations = relations

    def __repr__(self):
        return "MissingRelations: {}".format(", ".join(self.relations))



"""
Service class
------------------
Service object for interfacing with the Phenotype Archive API

"""

import os
import logging
from posixpath import join as urljoin
from requests import codes
from typing import Optional, List, Union, Dict

from ...client import Client
from ...services import BaseService
from ...exceptions import ServerError
from .exceptions import PhenotypeError
from .phenotype import Phenotype
from .playlist import Playlist
from .phenotype_matrix import PhenotypeMatrix
from .analysis_catalog import AnalysisCatalog
from .analysis_catalog_run import AnalysisCatalogRun

SERVICE_PATH = "api/phenotype-catalog"

log = logging.getLogger(__file__)

SUPPORTED_RESULT_TYPES = ["SET", "QT", "CATEGORY"]


def _get_paginated_results(method, limit):
    default_batch_size = 100  # Used for pagination
    offset = 0
    combined_data = []
    # Loop to fetch the entire results, combining the paginated results
    while True:
        batch_size = min(default_batch_size,
                         limit - offset)  # Make sure we dont get request too many phenos in last api call
        data = method(batch_size, offset)
        results = len(data)
        combined_data += data
        offset += results

        if results < batch_size or offset >= limit:  # Stop if response smaller than requested or results (larger or) equal to limit
            break
    return combined_data


class Service(BaseService):
    """
    A connection to the phenotype catalog service for a specific.

    The project can be passed in e.g. `nextcode.get_service("phenotype", project=myproject)`
    If it is not, the GOR_API_PROJECT environment variable or current profile config will be used.
    A project must be set when the service is instantiated.

    If the project does not already exist in the phenotype catalog only the `create_phenotype`
    method will work and the project will be created implicitly.

    To view available projects use the `svc.all_projects` dict
    """

    project_name: str = ""
    project: Dict = {}
    links: Dict = {}
    all_projects: Dict = {}

    def __init__(self, client: Client, *args, **kwargs) -> None:
        super(Service, self).__init__(client, SERVICE_PATH, *args, **kwargs)
        project_name = (
            kwargs.get("project")
            or os.environ.get("GOR_API_PROJECT")
            or client.profile.project
        )
        if not project_name:
            raise PhenotypeError("Please specify a project")

        self._init_project(project_name)

    def _init_project(self, project_name):
        """
        Initialize the project from the server
        """
        self.project_name = project_name
        resp = self.session.get(self.session.url_from_endpoint("projects"))
        self.all_projects = {}
        for project in resp.json()["projects"]:
            self.all_projects[project["name"].lower()] = project

        self.project = self.all_projects.get(self.project_name, {})
        self.links = self.project.get("links", {})
        if self.project:
            log.info(f"Service initialized with project {self.project_name}")
        else:
            log.warning(f"Project {self.project_name} not found")

    def create_phenotype(
            self,
            name: str,
            result_type: str,
            description: Optional[str] = None,
            url: Optional[str] = None,
            category: Optional[str] = None,
            query: Optional[str] = None,
            tags: Optional[List[str]] = [],
    ) -> Phenotype:
        """
        Create a new phenotype in the current project

        :param name: Unique (lowercase) phenotype name in the project
        :param result_type: Must be one of SET, QT or CATEGORY
        :param description: Free text description of the phenotype (optional)
        :param url: Reference URL for the phenotype (to dataset or other reference)
        :param category: Enter the category for the phenotype (must be defined in the project - see get_categories) (optional)
        :param query: NOR query that defines this phenotype (optional)
        :param tags: comma separated list of tags to add to this phenotype (optional) e.g. ['tag1','tag2']
        :raises: PhenotypeError, ServerError
        """
        uri = urljoin(
            self.session.url_from_endpoint("root"),
            "projects",
            self.project_name,
            "phenotypes",
        )
        result_type = result_type.upper()
        if result_type not in SUPPORTED_RESULT_TYPES:
            raise PhenotypeError(
                f"Result type {result_type} not supported. Use one of {', '.join(SUPPORTED_RESULT_TYPES)}"
            )
        payload = {
            "name": name,
            "result_type": result_type,
            "description": description,
            "url": url,
            "category": category,
            "query": query,
            "tag_list": tags,
        }
        resp = self.session.post(uri, json=payload)
        resp.raise_for_status()
        data = resp.json()

        # if the project did not already exist, initialize the service
        if not self.project:
            self._init_project(self.project_name)
        return Phenotype(self.session, data["phenotype"])

    def get_tags(self) -> List:
        """
        A list of all tags available in the system
        """
        resp = self.session.get(self.session.url_from_endpoint("tags"))
        return resp.json()["tags"]

    def get_phenotypes(
            self,
            all_tags: List[str] = [],
            any_tags: List[str] = [],
            categories: List[str] = [],
            limit: int = 100,
            states: List[str] = [],
            search: str = None,
            playlist: int = None,
            updated_at: str = None,
            result_types: List[str] = [],
            names: Optional[List[str]] = [],
            pn_count: Optional[str] = None
    ) -> List[Phenotype]:
        """
        Get all phenotypes in the current project as a list of Phenotypes.

        :param all_tags: Only fetch phenotypes that have all tags in the given list of tags
        :param any_tags: Fetch phenotypes that have any of the tags in the given list of tags
        :param categories: Only fetch phenotypes in the given list of categories
        :param limit: Maximum number of results (default: 100)
        :param states: Only fetch phenotypes in the given list of states
        :param search: String of keywords to search for in phenotypes, such as name, categories and tags
        :param playlist: Fetch a specific playlist of phenotypes by the playlist id
        :param updated_at: Only fetch phenotypes that match the given dates. Example: >=2017-04-01 ┃ <=2012-07-04 ┃ 2016-04-30..2016-07-04
        :param result_types: Only fetch phenotypes in the given list of result types
        :param names: Only fetch phenotypes in the given list of phenotype names
        :param pn_count: Only fetch phenotypes that match the given pn counts.
            "<10" matches records where the attribute less than 10.
            ">=10" matches records where the attribute is greater than or equal to 10.
            "=30" matches records where the attribute is equal to 30.
            "10..20" matches records where the attribute is between 10 and 20 (both included).
        :return: List of Phenotype
        :raises: `PhenotypeError` if the project does not exist
        :raises: ServerError
        """
        combined_data = self._get_phenotypes(
            all_tags,
            any_tags,
            categories,
            limit,
            states,
            search,
            playlist,
            updated_at,
            result_types,
            names,
            pn_count
        )
        phenotypes = []
        for item in combined_data:
            phenotypes.append(Phenotype(self.session, item))
        return phenotypes

    def get_phenotypes_matrix(
            self,
            all_tags: List[str] = [],
            any_tags: List[str] = [],
            categories: List[str] = [],
            limit: int = 100,
            states: List[str] = [],
            search: str = None,
            playlist: int = None,
            updated_at: str = None,
            result_types: List[str] = [],
            names: Optional[List[str]] = [],
            pn_count: Optional[str] = None
    ) -> List[Phenotype]:
        """
        Get all phenotypes in the current project as a PhenotypeMatrix.

        :param all_tags: Only fetch phenotypes that have all tags in the given list of tags
        :param any_tags: Fetch phenotypes that have any of the tags in the given list of tags
        :param categories: Only fetch phenotypes in the given list of categories
        :param limit: Maximum number of results (default: 100)
        :param states: Only fetch phenotypes in the given list of states
        :param search: String of keywords to search for in phenotypes, such as name, categories and tags
        :param playlist: Fetch a specific playlist of phenotypes by the playlist id
        :param updated_at: Only fetch phenotypes that match the given dates. Example: >=2017-04-01 ┃ <=2012-07-04 ┃ 2016-04-30..2016-07-04
        :param result_types: Only fetch phenotypes in the given list of result types
        :param names: Only fetch phenotypes in the given list of phenotype names
        :param pn_count: Only fetch phenotypes that match the given pn counts.
            "<10" matches records where the attribute less than 10.
            ">=10" matches records where the attribute is greater than or equal to 10.
            "=30" matches records where the attribute is equal to 30.
            "10..20" matches records where the attribute is between 10 and 20 (both included).
        :return: Phenotypes as PhenotypeMatrix
        :raises: `PhenotypeError` if the project does not exist
        :raises: ServerError
        """
        combined_data = self._get_phenotypes(
            all_tags,
            any_tags,
            categories,
            limit,
            states,
            search,
            playlist,
            updated_at,
            result_types,
            names,
            pn_count
        )
        matrix = PhenotypeMatrix(self.session, project_name = self.project_name)
        if combined_data:
            matrix.add_phenotypes([item['name'] for item in combined_data])
        return matrix

    def get_phenotypes_dataframe(
            self,
            all_tags: List[str] = [],
            any_tags: List[str] = [],
            categories: List[str] = [],
            limit: int = 100,
            states: List[str] = [],
            search: str = None,
            playlist: int = None,
            updated_at: str = None,
            result_types: List[str] = [],
            names: Optional[List[str]] = [],
            pn_count: Optional[str] = None,
            detail: Optional[bool] = False
    ) -> List[Phenotype]:
        """
        Get all phenotypes in the current project as a pandas DataFrame.

        :param all_tags: Only fetch phenotypes that have all tags in the given list of tags
        :param any_tags: Fetch phenotypes that have any of the tags in the given list of tags
        :param categories: Only fetch phenotypes in the given list of categories
        :param limit: Maximum number of results (default: 100)
        :param states: Only fetch phenotypes in the given list of states
        :param search: String of keywords to search for in phenotypes, such as name, categories and tags
        :param playlist: Fetch a specific playlist of phenotypes by the playlist id
        :param updated_at: Only fetch phenotypes that match the given dates. Example: >=2017-04-01 ┃ <=2012-07-04 ┃ 2016-04-30..2016-07-04
        :param result_types: Only fetch phenotypes in the given list of result types
        :param names: Only fetch phenotypes in the given list of phenotype names
        :param pn_count: Only fetch phenotypes that match the given pn counts.
            "<10" matches records where the attribute less than 10.
            ">=10" matches records where the attribute is greater than or equal to 10.
            "=30" matches records where the attribute is equal to 30.
            "10..20" matches records where the attribute is between 10 and 20 (both included).
        :param detail Return subset of phenotype info columns. detail=True returns all
        :return: Phenotypes as pandas dataframe
        :raises: `PhenotypeError` if the project does not exist
        :raises: ServerError
        """
        combined_data = self._get_phenotypes(
            all_tags,
            any_tags,
            categories,
            limit,
            states,
            search,
            playlist,
            updated_at,
            result_types,
            names,
            pn_count
        )
        import pandas
        dataframe = pandas.DataFrame(combined_data)

        if not dataframe.empty and not detail:
            cols = ["name", "description", "result_type", "tag_list", "pn_count"]
            return dataframe[cols]
        return dataframe

    def _get_phenotypes(
            self,
            all_tags: List[str] = [],
            any_tags: List[str] = [],
            categories: List[str] = [],
            limit: int = 100,
            states: List[str] = [],
            search: str = None,
            playlist: int = None,
            updated_at: str = None,
            result_types: List[str] = [],
            names: Optional[List[str]] = [],
            pn_count: Optional[str] = None
    ) -> List[Phenotype]:
        """
        Internal method to be called by `get_phenotypes`, `get_phenotypes_matrix` and `get_phenotypes_dataframe`
        See documentation of those methods for more details

        :param all_tags: Only fetch phenotypes that have all tags in the given list of tags
        :param any_tags: Fetch phenotypes that have any of the tags in the given list of tags
        :param categories: Only fetch phenotypes in the given list of categories
        :param limit: Maximum number of results (default: 100)
        :param states: Only fetch phenotypes in the given list of states
        :param search: String of keywords to search for in phenotypes, such as name, categories and tags
        :param playlist: Fetch a specific playlist of phenotypes by the playlist id
        :param updated_at: Only fetch phenotypes that match the given dates. Example: >=2017-04-01 ┃ <=2012-07-04 ┃ 2016-04-30..2016-07-04
        :param result_types: Only fetch phenotypes in the given list of result types
        :param names: Only fetch phenotypes in the given list of phenotype names
        :param pn_count: Only fetch phenotypes that match the given pn counts.
        :return: List of phenotypes
        :raises: `PhenotypeError` if the project does not exist
        :raises: ServerError
        """
        # TODO: project name is case sensitive on the server

        if not self.project:
            raise PhenotypeError("Project does not exist.")

        url = self.links["phenotypes"]
        if playlist:
            url = urljoin(self.links["self"], "playlists", str(playlist))

        if all_tags:
            all_tags = ','.join(all_tags) if isinstance(all_tags, list) else all_tags

        if any_tags:
            any_tags = ','.join(any_tags) if isinstance(any_tags, list) else any_tags

        if categories:
            categories = ','.join(categories) if isinstance(categories, list) else categories

        if result_types:
            result_types = ','.join(result_types) if isinstance(result_types, list) else result_types

        if states:
            states = ','.join(states) if isinstance(states, list) else states

        if names:
            names = ','.join(names)

        def do_get(batch_size, offset):
            # This local method fetches paginated results from `offset` to limit
            content = {
                "with_all_tags": all_tags,
                "with_any_tags": any_tags,
                "limit": batch_size,
                "offset": offset,
                "category": categories,
                "search": search,
                "state": states,
                "updated_at": updated_at,
                "result_type": result_types,
                "names": names,
                "pn_count": pn_count
            }
            resp = self.session.get(url, data={k: v for k, v in content.items() if v is not None})

            if playlist:
                data = resp.json()["playlist"]["phenotypes"]
            else:
                data = resp.json()["phenotypes"]
            return data

        combined_data = _get_paginated_results(do_get, limit)
        return combined_data

    def get_phenotype(self, name: str) -> Phenotype:
        """
        Get a specific phenotype in the current project

        :param name: Unique (lowercase) phenotype name in the project
        :return: List of phenotypes as per api spec
        :raises: `PhenotypeError` if the project does not exist
        :raises: `ServerError`
        """
        # TODO: project name is case sensitive on the server
        if not self.project:
            raise PhenotypeError("Project does not exist.")

        url = urljoin(self.links["phenotypes"], name)
        try:
            resp = self.session.get(url)
        except ServerError as ex:
            if ex.response and ex.response["code"] == codes.not_found:
                raise PhenotypeError("Phenotype not found") from None
            else:
                raise

        data = resp.json()["phenotype"]
        return Phenotype(self.session, data)

    def get_phenotype_matrix(self, base: Optional[str] = None) -> PhenotypeMatrix:
        """
        Get a phenotype matrix object.

        :param base: Optional name of base set
        :return: PhenotypeMatrix builder object
        :raises: `PhenotypeError` if the project does not exist
        :raises: `ServerError`
        """
        return PhenotypeMatrix(self.session, base=base, project_name=self.project_name)

    def get_categories(self) -> List:
        """
        A list of all categories available in the system
        """
        resp = self.session.get(
            urljoin(
                self.session.url_from_endpoint("root"),
                "projects",
                self.project_name,
                "categories",
            )
        )
        data = resp.json()["categories"]
        categories = []
        for item in data:
            categories.append(item)

        return categories

    def create_category(self, name: str):
        """
        Add a new category to this project.

        :param name: Name of the category
        :raises: `PhenotypeError` if the project does not exist
        """
        url = urljoin(
            self.session.url_from_endpoint("root"),
            "projects",
            self.project_name,
            "categories",
        )
        payload = {"name": name}
        resp = self.session.post(url, json=payload)
        resp.raise_for_status()
        data = resp.json()

        # if the project did not already exist, initialize the service
        if not self.project:
            self._init_project(self.project_name)

        return payload

    def create_playlist(self, name: str, description: Optional[str] = None) -> Playlist:
        """
        Create a new playlist in the current project

        :param name: Unique (lowercase) phenotype name in the project
        :param description: Free text description of the playlist (optional)
        :param phenotypes: comma separated list of phenotypes to add (optional) e.g. ['tag1','tag2']
        """

        url = urljoin(
            self.session.url_from_endpoint("projects"), self.project_name, "playlists"
        )
        payload = {"name": name, "description": description}
        resp = self.session.post(url, json=payload)
        resp.raise_for_status()
        data = resp.json()

        # if the project did not already exist, initialize the service
        if not self.project:
            self._init_project(self.project_name)
        return Playlist(self.session, data["playlist"])

    def get_playlists(self, limit: int = 100) -> List[Playlist]:
        """
        A list of all the playlists in the current project.

        :param limit: Maximum number of results (default: 100)
        :return: List of playlists
        :raises: `PhenotypeError` if the project does not exist
        :raises: ServerError
        """

        if not self.project:
            raise PhenotypeError("Project does not exist.")
        url = urljoin(
            self.session.url_from_endpoint("projects"), self.project_name, "playlists"
        )
        content = {"limit": limit}
        resp = self.session.get(url, data=content)

        data = resp.json()["playlists"]
        playlists = []
        for item in data:
            playlists.append(Playlist(self.session, item))
        return playlists

    def get_playlist(self, id: int = None, name: str = None) -> Playlist:
        """
        A list a single playlist in the current project based on the id.

        :param id: Specify the playlist to get by its id
        :param name: Retrieve a playlist by its unique name within project
        :return: A single playlist
        :raises: `PhenotypeError` if the project does not exist
        :raises: ServerError
        """

        if name and id:
            raise TypeError("id and name cannot both be supplied")
        if not self.project:
            raise PhenotypeError("Project does not exist.")
        url = urljoin(
            self.session.url_from_endpoint("projects"),
            self.project_name,
            "playlists",
            str(id),
        )

        if name:
            url = urljoin(
                self.session.url_from_endpoint('projects'),
                self.project_name,
                'playlists'
            )

        try:
            resp = self.session.get(url, data={'name': name})
        except ServerError as ex:
            if ex.response and ex.response["code"] == codes.not_found:
                raise PhenotypeError("Playlist not found") from None
            else:
                raise

        if name:
            data = resp.json()['playlists'][0]
        else:
            data = resp.json()["playlist"]
        return Playlist(self.session, data)

    def get_covariates(self, limit=100):
        """
        Get all covariates in current project
        """
        url = urljoin(self.links['self'], 'covariates')

        def do_get(batch_size, offset):
            content = {'limit': batch_size, 'offset': offset}
            resp = self.session.get(url, data=content)
            data = resp.json()['covariates']
            return data

        combined_data = _get_paginated_results(do_get, limit)

        return combined_data

    def get_covariate(self, id):
        """
        Get a single covariate by its id
        """
        url = urljoin(self.links['self'], 'covariates', str(id))
        try:
            resp = self.session.get(url)
        except ServerError as ex:
            if ex.response and ex.response['code'] == codes.not_found:
                raise PhenotypeError(f"Covariate not found") from None
            else:
                raise ex
        data = resp.json()['covariate']
        return data

    def get_analysis_catalogs(self, phenotype_name: Optional[str] = None, limit: int = 100) -> List[AnalysisCatalog]:
        """
        A list of all the analysis catalogs in the current project. Optionally scope results to a given phenotype name.

        :param phenotype_name: Only list analysis catalogs for a specific phenotype name (optional)
        :param limit: Maximum number of results (default: 100)
        :return: List of analysis catalogs
        :raises: `PhenotypeError` if the project does not exist
        :raises: ServerError
        """

        if not self.project:
            raise PhenotypeError("Project does not exist.")
        url = urljoin(
            self.session.url_from_endpoint("projects"), self.project_name, "analysis_catalogs"
        )

        if phenotype_name:
            url = urljoin(
                self.session.url_from_endpoint("projects"), self.project_name, "phenotypes", phenotype_name,
                "analysis_catalogs"
            )

        content = {"limit": limit}
        resp = self.session.get(url, data=content)

        data = resp.json()["analysis_catalogs"]
        analysis_catalogs = []
        for item in data:
            analysis_catalogs.append(AnalysisCatalog(self.session, item))
        return analysis_catalogs

    def get_analysis_catalog(self, analysis_catalog_name: str) -> AnalysisCatalog:
        """
        Get an analysis catalog in the current project by name

        :param analysis_catalog_name: The name of the Analysis Catalogs
        """

        url = urljoin(
            self.session.url_from_endpoint("projects"), self.project_name, "analysis_catalogs", analysis_catalog_name
        )

        resp = self.session.get(url)
        resp.raise_for_status()
        data = resp.json()

        # if the project did not already exist, initialize the service
        if not self.project:
            self._init_project(self.project_name)
        return AnalysisCatalog(self.session, data["analysis_catalog"])

    def create_analysis_catalog(
            self,
            playlist_id: str,
            name: str,
            recipe_name: str,
            recipe_parameters: Dict,
            covariate_phenotypes: Optional[List[str]] = [],
            excluded_pns: Optional[List[str]] = []
    ) -> AnalysisCatalog:
        """
        Create a new analysis catalog in the current project

        :param playlist_id: The id of the playlist that contains the phenotypes to use for analysis
        :param name: Analysis Catalog name
        :param recipe_name: The name of the recipe to use
        :param recipe_parameters: The parameters required to run the recipe
        :param covariate_phenotypes: The names of phenotypes to use as covariates, e.g. ['Pheno1','Pheno2'] (optional)
        :param excluded_pns: the PNs to exclude from the analysis, e.g. ['PN1','PN2'] (optional)
        """

        url = urljoin(
            self.session.url_from_endpoint("projects"), self.project_name, "playlists", playlist_id, "analysis_catalogs"
        )
        payload = {
            "analysis_catalog": {
                "name": name,
                "recipe_name": recipe_name,
                "recipe_parameters": recipe_parameters,
                "covariate_phenotypes": covariate_phenotypes,
                "excluded_pns": excluded_pns
            }
        }
        resp = self.session.post(url, json=payload)
        resp.raise_for_status()
        data = resp.json()

        # if the project did not already exist, initialize the service
        if not self.project:
            self._init_project(self.project_name)
        return AnalysisCatalog(self.session, data["analysis_catalog"])

    def get_analysis_catalog_run(self, analysis_catalog_name: str,
                                 analysis_catalog_run_name: str) -> AnalysisCatalogRun:
        """
        Get an analysis catalog run in the current project by names

        :param analysis_catalog_name: The name of the Analysis Catalog
        :param analysis_catalog_run_name: The name of the Analysis Catalog run
        """

        url = urljoin(
            self.session.url_from_endpoint("projects"), self.project_name, "analysis_catalogs", analysis_catalog_name,
            "runs", analysis_catalog_run_name
        )

        resp = self.session.get(url)
        resp.raise_for_status()
        data = resp.json()

        # if the project did not already exist, initialize the service
        if not self.project:
            self._init_project(self.project_name)
        return AnalysisCatalogRun(self.session, data["analysis_catalog_run"])

    def get_analysis_catalog_runs(self, phenotype_name: str, limit: int = 100) -> List[AnalysisCatalogRun]:
        """
        A list of all the analysis catalog runs in the current project for a given phenotype.

        :param phenotype_name: Only list analysis catalog runs for a specific phenotype name
        :param limit: Maximum number of results (default: 100)
        :return: List of analysis catalog runs
        :raises: `PhenotypeError` if the project does not exist
        :raises: ServerError
        """

        if not self.project:
            raise PhenotypeError("Project does not exist.")

        url = urljoin(
            self.session.url_from_endpoint("projects"), self.project_name, "phenotypes", phenotype_name,
            "analysis_catalog_runs"
        )

        content = {"limit": limit}
        resp = self.session.get(url, data=content)

        data = resp.json()["analysis_catalog_runs"]
        analysis_catalog_runs = []
        for item in data:
            analysis_catalog_runs.append(AnalysisCatalogRun(self.session, item))
        return analysis_catalog_runs



"""
client
~~~~~~~~~~

This is the basic entrypoint to interact with services.
"""

import os
import pkgutil
from importlib import import_module
from pathlib import Path
import logging
from typing import Optional, Union, Dict, List
from urllib.parse import urlsplit
from posixpath import join as urljoin
import requests
import json

from .exceptions import ServiceNotFound, InvalidProfile, InvalidToken
from .config import Config, save_config
from .utils import root_url_from_api_key, get_access_token, decode_token, host_from_url

SERVICES_PATH = Path(__file__).parent.joinpath("services")
SERVICES = ["query", "queryserver"]
DEFAULT_CLIENT_ID = "api-key-client"

log = logging.getLogger()

cfg = Config()


def get_api_key(
    host: str, username: str, password: str, realm: str = "wuxinextcode.com"
) -> str:
    """
    Authenticate with keycloak server and return an API key.

    Assumes client api-key-client.

    :param host: The URI of the service. e.g. host.wuxinextcode.com
    :param username: Username of the keycloak user which is authenticating
    :param password: Password
    :param realm: The realm with which to authenticate (optional)
    :returns: API Key which can be used in subsequent calls to Client()
    :raises: InvalidToken
    """
    body = {
        "grant_type": "password",
        "client_id": DEFAULT_CLIENT_ID,
        "password": password,
        "username": username,
        "scope": "offline_access",
    }
    host = host_from_url(host)
    url = urljoin(host, "auth", "realms", realm, "protocol/openid-connect/token")
    log.info("Using auth server '%s'", url)
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    log.debug("Calling POST %s with headers %s and body %s", url, headers, body)
    resp = requests.post(url, headers=headers, data=body)
    log.debug("Response (%s): %s", resp.status_code, resp.text)
    if resp.status_code != 200:
        try:
            description = resp.json()["error_description"]
        except Exception:
            description = resp.text
        raise InvalidToken(f"Error logging in: {description}") from None

    api_key = resp.json()["refresh_token"]
    return api_key


def get_service(
    service_name: str,
    api_key: Optional[str] = None,
    profile: Optional[str] = None,
    **kw,
):
    """
    Helper method which returns a service handle and creates a client object automatically

    :param api_key: api key to use for this client
    :param profile: name of a saved profile to use with this client
    :raises: InvalidProfile, ServiceNotFound
    """
    client = Client(api_key, profile)
    return client.service(service_name, **kw)


class Profile:

    profile_name = None
    content = None

    def __init__(
        self, profile_name: Optional[str] = None, content: Optional[Dict] = None
    ) -> None:
        if profile_name:
            self.profile_name = profile_name
            self.content = cfg.get("profiles")[self.profile_name]
        else:
            self.profile_name = None
            self.content = content

    def __getattr__(self, name):
        if name in self.content:
            return self.content[name]
        return None

    def __setattr__(self, name, value):
        if name in ("profile_name", "content"):
            super(Profile, self).__setattr__(name, value)
        else:
            self.content[name] = value
            if self.profile_name:
                save_config()


class Client:
    """A base client object from which to access various services

    :param api_key: api key to use for this client
    :param profile: name of a saved profile to use with this client
    :param root_url: override the URL of the server root. e.g. https://server.wuxinextcode.com
    :raises: InvalidProfile

    Order of precedence:
    1) passed in 'profile'
    2) passed in 'api_key'
    3) NEXTCODE_PROFILE from environment
    3) GOR_API_KEY from environment"""

    def __init__(
        self,
        api_key: Optional[str] = None,
        profile: Optional[str] = None,
        root_url: Optional[str] = None,
    ) -> None:
        # if no named profile or api key is passed in
        if not profile and not api_key:
            # find the default profile, if any
            if os.environ.get("NEXTCODE_PROFILE"):
                profile = os.environ["NEXTCODE_PROFILE"]
                log.info("Using profile '%s' from environment", profile)
            else:
                profile = cfg.get("default_profile")
                if profile:
                    log.info("Using default profile '%s' from config", profile)

        self.profile_name = profile

        # if we have a profile we will load it from the config
        if self.profile_name:
            log.info(
                "Initializing client with profile '%s'. Available profiles: %s",
                self.profile_name,
                self.available_profiles(),
            )
            if self.profile_name not in cfg.get("profiles"):
                raise InvalidProfile(
                    "The config profile (%s) could not be found" % profile
                )
            self.profile = Profile(self.profile_name)
        else:
            # if there is no profile there needs to be configuration set in the
            # environment, in which case we create an ephemeral profile, not
            # backed up by disk.
            api_key = api_key or os.environ.get("GOR_API_KEY")
            #if not api_key:
            #    raise InvalidProfile(
            #        "No profile specified and GOR_API_KEY not set in environment"
            #    )
            root_url = root_url or os.environ.get("NEXTCODE_ROOT_URL")
            if not root_url:
                root_url = root_url_from_api_key(api_key)
            self.profile = Profile(content={"api_key": api_key, "root_url": root_url})

    def service(self, service_name: str, **kw):
        """
        Retrieve an instance to a service.

        :param service_name: The name of the service

        Available services can be listed by calling class method `available_services`.
        """
        for (_, name, _) in pkgutil.iter_modules([str(SERVICES_PATH)]):
            if name == service_name:
                log.debug("Importing service %s", name)
                module = import_module("..services.{}".format(name), __name__)
                svc = module.Service(client=self, **kw)  # type: ignore
                return svc
        raise ServiceNotFound()

    def get_access_token(self, decode: bool = False) -> Union[Dict, str]:
        """Retrieve the JWT access token that is generated from the current api key.

        :param decode: Decode the token and return a dictionary instead of the raw string
        :raises: InvalidToken

        Example usage:

        >>> client = nextcode.Client(api_key="xxx.yyy.zzz")
        >>> print(client.get_access_token())
        "iii.jjj.kkk"

        >>> client = nextcode.Client(api_key="xxx.yyy.zzz")
        >>> print(client.get_access_token(decode=True))
        {"jti": "...", ...}

        """
        token = os.environ.get('NEXTCODE_ACCESS_TOKEN') or get_access_token(self.profile.api_key)
        if decode:
            return decode_token(token)
        else:
            return token

    @classmethod
    def available_services(cls) -> List[str]:
        """List services that can be intantiated via an client object client.service(`service`)"""
        ret = []
        for (_, name, _) in pkgutil.iter_modules([str(SERVICES_PATH)]):
            ret.append(name)
        return ret

    @classmethod
    def available_profiles(cls) -> List[str]:
        """List profiles that that are installed into the system and can be used
        with the constructor client.Client()
        """
        return list(cfg.get("profiles"))

    def __repr__(self) -> str:
        return "<Client {} | {}>".format(self.profile_name, self.profile.root_url)



"""
Workflow Job
------------------

The Query class represents a workflow job model from the RESTFul Workflow Service API

"""
import configparser
import json
import datetime
import os

import botocore.session
import dateutil
import time
import logging
from typing import Callable, Union, Optional, Dict, List

from . import RUNNING_STATUSES, FINISHED_STATUSES, FAILED_STATUSES
from .exceptions import JobError
from ...exceptions import ServerError
from ...session import ServiceSession

log = logging.getLogger(__name__)


class WorkflowJob:
    """
    Proxy object for a serverside workflow job.

    This object can be queried for the current status of a workflow job and will
    automatically refresh from server until the job is finished (or failed)
    """

    def __init__(self, session: ServiceSession, job_id: int, job: Dict):
        self.session = session
        self.job = job
        self.job_id = self.job["job_id"]
        self.links = self.job["links"]

    def __repr__(self) -> str:
        return json.dumps(self.job)

    @property
    def duration(self) -> str:
        """
        Stringified duration of job for human consumption
        """
        if self.complete_date:
            complete_date = self.complete_date
        elif self.status in RUNNING_STATUSES:
            complete_date = datetime.datetime.utcnow()
        elif self.status_date:
            complete_date = self.status_date
        else:
            return "-"
        ret = complete_date - self.submit_date
        # remove microseconds since no one wants them
        ret = ret - datetime.timedelta(microseconds=ret.microseconds)
        return ret

    @property
    def finished(self) -> bool:
        """
        Has the job finished (might be failed)
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in FINISHED_STATUSES

    @property
    def running(self) -> bool:
        """
        Is the job currently running
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in RUNNING_STATUSES

    @property
    def failed(self) -> bool:
        """
        Is the job in a failed state
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in FAILED_STATUSES

    @property
    def done(self) -> bool:
        """
        Has the job completed successfully
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status == "COMPLETED"

    def refresh(self) -> None:
        """
        Refresh the local cache of the serverside job object
        """
        self.job = self.session.get(self.links["self"]).json()

    def resume(self) -> None:
        """
        Rerun a job that has previously failed
        """
        _ = self.session.put(self.links["self"])
        self.refresh()

    def wait(self, max_seconds: Optional[int] = None, poll_period: float = 0.5):
        """
        Wait for a running job to complete.

        This is similar to the wait method in the Query Service and will wait by default
        indefinitely for a workflow job to complete and poll the server regularly to update
        the local status. When the job is completed (or failed) the method will return the
        workflow job object.

        :param max_seconds: raise an exception if the job runs longer than this
        :param poll_period: Number of seconds to wait between polling (max 10 seconds)
        :returns: WorkflowJob
        :raises: :exc:`JobError`
        """
        if not self.running:
            return self
        log.info("Waiting for job %s to complete...", self.job_id)
        start_time = time.time()
        duration = 0.0
        period = poll_period
        is_running = self.running
        while is_running:
            time.sleep(period)
            duration = time.time() - start_time
            is_running = self.running

            # cancel the wait if the executor pod is in trouble after 30 seconds of waiting to start.
            # it most likely means that the nextflow script has a syntax error or something.
            if self.status == "PENDING" and (
                max_seconds and duration > max_seconds or duration > 30.0
            ):
                log.info("Job has been PENDING for %.0fsec. Inspecting it...", duration)
                curr_status = self.inspect()
                executor_pod = None
                for pod in curr_status["pods"]:
                    if pod["metadata"]["labels"]["app-name"] == "nextflow-executor":
                        executor_pod = pod
                        break
                if not executor_pod:
                    log.warning(
                        "Job is still pending after %.0fsec and executor pod is missing",
                        duration,
                    )
                    raise JobError(
                        f"Job is still pending after {duration:.0f}s and executor pod is missing. View logs or inspect job for failures."
                    )
                if not executor_pod["status"]["container_statuses"][0]["state"][
                    "running"
                ]:
                    log.warning(
                        "Job is still pending after %.0fsec and executor pod is not in running state",
                        duration,
                    )
                    raise JobError(
                        f"Job is still pending after {duration:.0f}s and executor pod is not in running state. View logs or inspect job for failures."
                    )

            if is_running and max_seconds and duration > max_seconds:
                raise JobError(
                    f"Job {self.job_id} has exceeded wait time {max_seconds}s and we will not wait any longer. It is currently {self.status}."
                )

            period = min(period + 0.5, 10.0)
        if self.status == "DONE":
            log.info(
                "Job %s completed in %.2f sec and returned %s rows",
                self.job_id,
                duration,
                self.line_count,
            )
        else:
            log.info(
                "Job %s has status %s after %.2f sec",
                self.job_id,
                self.status,
                duration,
            )
        return self

    def cancel(
        self, status: Optional[str] = None, status_message: Optional[str] = None
    ) -> Optional[str]:
        """
        Cancel a running job

        :param status: status of the cancelled job. Defaults to CANCELLED
        :returns: Status message from the server
        """
        data: Dict = {}
        data["status"] = status or "CANCELLED"
        data["status_message"] = status_message
        resp = self.session.delete(self.links["self"], json=data)
        status_message = resp.json()["status_message"]
        return status_message

    def inspect(self) -> Dict:
        """
        Inspect a failed job for debugging.

        Returns unfiltered pod and node information from the kubernetes system

        :returns: Dictionary containing low-level debugging information
        :raises: :exc:`ServerError` if the server is not configured for inspection capabilities
        """
        try:
            url = self.links["inspect"]
        except KeyError:
            raise ServerError("Server does not support inspect functionality")
        resp = self.session.get(url)
        return resp.json()

    def cost(self, recalculate: bool = False) -> Dict:
        """
        Get the estimates cost of this job.
        :param recalculate: If set to True then the cost will be recalculated.
        """
        url = self.links["cost"]
        data = {}
        if recalculate:
            data["recalculate"] = "true"

        resp = self.session.get(url, json=data)
        return resp.json()

    def processes(
        self,
        process_id: Optional[int] = None,
        is_all: bool = False,
        limit: int = 50,
        status: Optional[str] = None,
    ) -> List[Dict]:
        """
        Get a list of all nextflow processes in this job

        :param process_id: process_id to show
        :param is_all: Show all processes, otherwise show only running processes
        :param limit: Maximum number of processes to return
        :param status: Filter processes by status
        """
        url = self.links["processes"]
        if process_id:
            url += "/%s" % process_id
            resp = self.session.get(url)
            return [resp.json()]
        else:
            data: Dict = {"limit": limit}
            if is_all:
                data["all"] = 1
            if status:
                data["status"] = status
            resp = self.session.get(url, json=data)
            return resp.json()["processes"]

    def events(self, limit: int = 50) -> List[Dict]:
        """
        Get a list of events reported by Nextflow for this job

        :param limit: Maximum number of events to return
        """
        url = self.links["events"]
        data = {"limit": limit}
        resp = self.session.get(url, json=data)
        return resp.json()["events"]

    def log_groups(self) -> Dict:
        """
        Get available log groups

        :returns: Dictionary with log group name and server url to the log group
        """
        logs_url = self.links["logs"]
        resp = self.session.get(logs_url)
        return resp.json()["links"]

    def logs(self, log_group: str = "pod", log_filter: Optional[str] = None) -> str:
        """
        Get text logs for the specified log group

        :param log_group: Name of the log group to view
        :param log_filter: Optional filter to apply to the logs
        :raises: :exc:`ServerError` If the log group is not available
        """
        groups = self.log_groups()
        url = None
        for k, v in groups.items():
            if k.startswith(log_group):
                url = v
                break
        if not url:
            raise ServerError(f"Log Group '{log_group}' is not available.")
        if log_filter:
            url += "?filter=%s" % log_filter
        logs = self.session.get(url).text
        return logs

    def __getattr__(self, name):
        try:
            val = self.job[name]
        except KeyError:
            raise AttributeError

        if name.endswith("_date") and val:
            val = dateutil.parser.parse(val)
        return val



"""
csa
~~~~~~~~~~
CSA management features
"""

from posixpath import join as urljoin
import requests
from requests import codes
import logging

from .exceptions import AuthServerError, CSAError
from .utils import host_from_url

log = logging.getLogger(__name__)


def _get_csa_error(resp):
    try:
        msg = resp.json()["error"]["full_message"]
        if isinstance(msg, list):
            msg = ", ".join(msg)
    except Exception:
        msg = str(resp.content)
    return msg


def _check_csa_error(resp):
    if 400 <= resp.status_code < 500:
        raise CSAError(_get_csa_error(resp))
    resp.raise_for_status()


class CSASession:
    def __init__(self, root_url, user_name, password):
        self.root_url = host_from_url(root_url)
        self.session = requests.Session()
        self.session.auth = (user_name, password)
        self.csa_url = urljoin(self.root_url, "csa/api/")

        resp = self.session.get(urljoin(self.csa_url, "users.json"), timeout=2.0)

        if resp.status_code == codes.unauthorized:
            raise AuthServerError(
                f"User {user_name} could not authenticate with CSA Server"
            )
        resp.raise_for_status()

    def get_user_key(self, user_name):
        users_url = urljoin(self.csa_url, "users.json")
        resp = self.session.get(users_url)

        resp.raise_for_status()
        users = resp.json()["users"]
        for user in users:
            if user["email"] == user_name:
                return user["key"]
        raise AuthServerError(f"User {user_name} not found")

    def create_user(self, user_name, password, exist_ok=False):
        try:
            _ = self.get_user_key(user_name)
        except AuthServerError:
            pass
        else:
            log.info("User '%s' already exists in CSA", user_name)
            if not exist_ok:
                raise CSAError(f"User '{user_name}' already exists.")
            else:
                return

        users_url = urljoin(self.csa_url, "users.json")
        payload = {"user": {"email": user_name, "password": password}}
        resp = self.session.post(users_url, json=payload)
        _check_csa_error(resp)
        log.info("Created user '%s' in CSA", user_name)

    def add_user_to_project(
        self, user_name, project, role="researcher", exist_ok=False
    ):
        user_key = self.get_user_key(user_name)
        roles_url = urljoin(self.csa_url, "user_roles.json")
        resp = self.session.post(
            roles_url,
            json={
                "user_role": {
                    "project_key": project,
                    "role": role,
                    "user_key": user_key,
                }
            },
        )
        if not exist_ok:
            _check_csa_error(resp)
        log.info(
            "User '%s' has been added with role %s to project %s",
            user_name,
            role,
            project,
        )

    def get_project_names(self):
        projects_url = urljoin(self.csa_url, "projects.json")
        resp = self.session.get(projects_url)
        resp.raise_for_status()
        projects = resp.json()["projects"]
        return [p["key"] for p in projects]

    def get_project(self, project_name):
        url = urljoin(self.csa_url, f"projects/{project_name}.json")
        resp = self.session.get(url)
        if resp.status_code == codes.not_found:
            return None
        resp.raise_for_status()
        return resp.json()["project"]

    def create_project(
        self, project_name, org_name, ref_version="hg38", exist_ok=False
    ):
        existing_project = self.get_project(project_name)
        if existing_project:
            if not exist_ok:
                raise CSAError("Project already exists")
            else:
                return existing_project

        url = urljoin(self.csa_url, "projects.json")
        data = {
            "project": {
                "key": project_name,
                "name": project_name,
                "organization_key": org_name,
                "reference_version": ref_version,
            }
        }
        resp = self.session.post(url, json=data)
        _check_csa_error(resp)
        return resp.json()["project"]

    def add_credentials(self, owner_key, service, lookup_key, credential_attributes):
        lookup_key = lookup_key.lower()
        url = urljoin(self.csa_url, "auth/v1/credentials.json")
        url = url.replace("/api", "")

        log.info(
            "Calling '%s' to add '%s' credentials to CSA project '%s'",
            url,
            service,
            owner_key,
        )
        resp = self.session.post(
            url,
            json={
                "credential": {
                    "owner_type": "Project",
                    "owner_key": owner_key,
                    "service": service,
                    "lookup_key": lookup_key,
                    "expires": "",
                    "credential_attributes": credential_attributes,
                }
            },
        )
        _check_csa_error(resp)
        return resp.json()

    def add_s3_credentials(self, owner_key, bucket_name, aws_key, aws_secret):
        return self.add_credentials(
            owner_key, "s3", bucket_name, {"key": aws_key, "secret": aws_secret}
        )

    def add_s3_credentials_to_project(
        self, project_name, bucket_name, aws_key, aws_secret
    ):
        return self.add_s3_credentials(project_name, bucket_name, aws_key, aws_secret)

    def get_s3_credentials_for_project(self, project_name):
        url = urljoin(
            self.csa_url,
            f"auth/v1/credentials.json?find[service]=s3&find[project_key]={project_name}",
        )
        url = url.replace("/api", "")
        resp = self.session.get(url)
        resp.raise_for_status()
        creds = resp.json()["credentials"]
        ret = {}
        for cred in creds:
            ret[cred["lookup_key"]] = {
                "aws_access_key_id": cred["credential_attributes"]["key"],
                "aws_secret_access_key": cred["credential_attributes"]["secret"],
                "project_name": cred["owner_key"],
            }
        return ret

    def import_sample(self, sample_data):
        url = urljoin(self.csa_url, "sample_data_sets.json")
        json_data = {"sample_data_set": sample_data}
        log.info("Importing sample into CSA")
        log.debug("Calling '%s' to import sample: %s", url, json_data)
        resp = self.session.post(url, json=json_data)
        _check_csa_error(resp)
        log.debug("Import sample returned content '%s'", resp.content)
        return resp.json()



"""
Query result object for Query Server queries.
"""
import json
import logging
import time
import zlib
from typing import Optional
from io import StringIO

from requests import Response

from nextcode.exceptions import ServerError
from nextcode.services.query.exceptions import QueryError, MissingRelations
from nextcode.utils import jupyter_available

log = logging.getLogger(__name__)


class Result(object):
    """
    Query result object for Query Server queries.
    """
    def __init__(self, resp: Response, gzip: bool):
        self.resp = resp
        self.open = True
        self.start_time = time.time()
        self.num_bytes = 0
        self.num_lines = -1  # For the header.
        self.gzip = gzip

    def iter_lines(self, limit: Optional[int] = None, callback=None):
        """
        Return iterator to iterate through the result.
        :param limit: Maximum number of rows to return (default all)
        :param callback: callback for progress
        :return: iterator
        """
        self.__check_open__()
        self.start_time = time.time()
        log.info(f"Starting to stream lines...")
        for line in self.iter_lines_unzip(decode_unicode=False, delimiter=b'\n'):

            if line:
                if line.startswith(b'#> EXCEPTION'):
                    throw_error_from_line(line.decode('utf-8'))
                elif line.startswith(b'#>'):
                    continue

                self.num_lines += 1
                self.num_bytes += len(line)

                if self.num_lines % 10000 == 0:
                    _log_download_progress(
                        self.num_lines / 10000,
                        self.num_bytes,
                        self.num_lines,
                        self.num_lines,
                        1,
                        self.start_time,
                        callback,
                    )

                if limit is not None and self.num_lines > limit:
                    break

                yield line.decode('utf-8')
        self.__close_response__()

    def lines(self, limit: Optional[int] = None):
        """
        Get query result lines in one string.
        :param limit: Maximum number of rows to return (default all).
        :return: query result lines in one string.
        """
        result_text: str = ""
        for line in self.iter_lines(limit):
            if len(result_text) > 0:
                result_text += '\n'
            result_text += line
        return result_text

    def content(self):
        """Content of the result, in bytes."""
        self.__check_open__()
        content = self.resp.content
        self.__close_response__()
        return content

    def text(self):
        """Content of the result, in unicode."""
        self.__check_open__()
        text = self.resp.text
        self.__close_response__()
        return text

    def dataframe(self, limit: Optional[int] = None):
        """
        Return a Pandas dataframe object containing the results of this query.

        :param limit: Maximum number of rows to return (default all).
        :raises QueryError: If the pandas library is not installed.
        :return: Pandas dataframe object.
        """
        if not jupyter_available():
            raise QueryError("Pandas library is not installed")
        tsv_data = self.lines(limit=limit)

        import pandas as pd
        if not tsv_data:
            return pd.DataFrame()
        df = pd.read_csv(StringIO(tsv_data), delimiter="\t")  # type: ignore
        return df

    def cancel(self):
        self.__close_response__()

    def __close_response__(self):
        self.resp.close()
        self.open = False

    def __check_open__(self):
        if not self.open:
            raise Exception("Response has been closed, data can not been accessed")

    ITER_CHUNK_SIZE = 512

    def iter_lines_unzip(self, chunk_size=ITER_CHUNK_SIZE, decode_unicode=False, delimiter=None):
        """Iterates over the response data, one line at a time.  When
        stream=True is set on the request, this avoids reading the
        content at once into memory for large responses.

        .. note:: This method is not reentrant safe.
        """

        pending = None

        d = zlib.decompressobj(16+zlib.MAX_WBITS)

        for chunk in self.resp.iter_content(chunk_size=chunk_size, decode_unicode=decode_unicode):

            if self.gzip:
                chunk = d.decompress(chunk)

            if pending is not None:
                chunk = pending + chunk

            if delimiter:
                lines = chunk.split(delimiter)
            else:
                lines = chunk.splitlines()

            if lines and lines[-1] and chunk and lines[-1][-1] == chunk[-1]:
                pending = lines.pop()
            else:
                pending = None

            for line in lines:
                yield line

        if pending is not None:
            yield pending


def _log_download_progress(
        num_chunk,
        num_bytes,
        num_lines,
        total_received_lines,
        total_expected_lines,
        start_time,
        callback=None,
):
    """
    Helper for logging out download progress and notifying a listener or delta lines
    """
    mb = num_bytes / 1024 / 1024
    if callback:
        callback(num_lines)
    diff = time.time() - start_time
    if num_chunk is not None and num_chunk % 1000 == 0:
        msg = f"Downloaded {total_received_lines}/{total_expected_lines} lines ({mb:.2f} MB) in {diff:.2f} sec"
        log.info(msg)


def throw_error_from_line(line):
    if line.startswith("#> EXCEPTION") and line.find('{') > -1:
        # Assume this is an exception from Query Server with embedded json.
        json_ex = json.loads(line[line.find('{'):line.rfind('}')+1])
        if "errorType" in json_ex and json_ex["errorType"] == "GorMissingRelationException":
            raise MissingRelations(
                [r for r in json_ex["uri"].split(',')]
            )
        else:
            raise ServerError(json_ex["gorMessage"])
    else:
        raise ServerError("Error while running query.\n{}".format(line))



"""
Service class
------------------
Service object for interfacing with the Project service API

"""

import logging

from ...client import Client
from ...services import BaseService

from .service import Service



"""
Playlist
------------------

A playlist is a collection of phenotypes that have 
been grouped together by a user in the Phenotype Catalog.


"""

import json
import datetime
import dateutil
import time
import logging
from posixpath import join as urljoin
from typing import Callable, Union, Optional, Dict, List

from .exceptions import PhenotypeError
from .phenotype_matrix import PhenotypeMatrix
from ...exceptions import ServerError
from ...session import ServiceSession

log = logging.getLogger(__name__)


class Playlist:
    """
    A local object representing a playlist response from the phenotype
    catalog service.

    Please refer to the API Documentation for the phenotype catalog service.

    In addition to the ones documented here, this object has at least these attributes:

    * name - Playlist name
    * description - Textual description of this phenotype
    * created_at - Timestamp when the playlist was first created
    * updated_at - Timestamp when the playlist was last updated
    * created_by - Username who created the phenotype
    * versions - List of data versions available in this phenotype

    """

    def __init__(self, session: ServiceSession, data: Dict):
        self.session = session
        self.data = data
        self.links = data["links"]

    def __getattr__(self, name):
        try:
            val = self.data[name]
        except KeyError:
            raise AttributeError

        try:
            val = dateutil.parser.parse(val)
        except Exception:
            pass

        return val

    def __repr__(self) -> str:
        return f"<Playlist {self.name} in project {self.project_key}>"

    def delete(self):
        """
        Delete a playlist from a project

        :raises: `ServerError` if the phenotype could not be deleted
        """
        _ = self.session.delete(self.links["self"])

    def add_phenotype(self, name: str):
        """
        Add a phenotype to a playlist
        """
        url = urljoin(self.links["self"], "phenotypes")
        content = {"name": name}
        _ = self.session.post(url, json=content)

        self.get_info()

    def get_info(self):
        return self.data

    def list_phenotypes(self):
        return [phenotype['name'] for phenotype in self.get_info()['phenotypes']]
    
    def get_data(self, missing_value=None):
        """
        Retrieve phenotype data from the server.
        """
        matrix = PhenotypeMatrix(self.session, project_name = self.data["project_key"])
        matrix.add_phenotypes(names=self.list_phenotypes(), missing_value=missing_value)
        self.df = matrix.get_data()
        return self.df



"""
Phenotype
------------------

Representation of a serverside phenotype.


"""

import json
import datetime
import dateutil
import time
import logging
from typing import Callable, Union, Optional, Dict, List
try:
    import pandas as pd
except ModuleNotFoundError:
    print('pandas is not installed - some functions might not work')
try:
    import plotly.graph_objects as go
    from plotly.offline import init_notebook_mode, iplot
except ModuleNotFoundError:
    print('plotly is not installed - some functions might not work')

from .exceptions import PhenotypeError
from .phenotype_matrix import PhenotypeMatrix
from ...exceptions import ServerError
from ...session import ServiceSession

log = logging.getLogger(__name__)

class Phenotype:
    """
    A local object representing a phenotype response from the phenotype
    catalog service.

    Note that most of the attributes come directly from the phenotype
    serverside response and are therefore not documented directly here.
    Please refer to the API Documentation for the phenotype catalog service.

    In addition to the ones documented here, this object has at least these attributes:

    * name - Phenotype name
    * description - Textual description of this phenotype
    * result_type - Type of result. Cannot be changed. One of SET, QT, CATEGORY
    * created_at - Timestamp when the phenotype was first created
    * updated_at - Timestamp when the phenotype was last updated
    * created_by - Username who created the phenotype
    * versions - List of data versions available in this phenotype

    To intraspect the data you can call `phenotype.data.keys()`. Each key
    in the `data` dict is exposed as an attribute with intelligent type casting.

    """

    def __init__(self, session: ServiceSession, data: Dict):
        self.session = session
        self.data = data
        self.links = data["links"]
        self.df = None

    def __getattr__(self, name):
        try:
            val = self.data[name]
        except KeyError:
            raise AttributeError

        try:
            val = dateutil.parser.parse(val)
        except Exception:
            pass

        return val

    def __repr__(self) -> str:
        return f"<Phenotype {self.name} in project {self.project_key}>"

    def refresh(self):
        """
        Refresh the local cache
        """
        resp = self.session.get(self.links["self"])
        data = resp.json()
        self.data = data["phenotype"]

    def delete(self):
        """
        Delete a phenotype, including all data from a project

        :raises: `ServerError` if the phenotype could not be deleted
        """
        _ = self.session.delete(self.links["self"])

    def upload(self, data: Union[List,pd.DataFrame]):
        """
        Upload phenotype data

        The data is expected to be either a list of lists or pandas DataFrame.
        e.g. `phenotype.upload([['a'], ['b']])`.
        The `result_type` of the phenotype dictates
        if each sublist should contain one or two items.

        :raises: `ServerError` if there was a problem uploading
        """
        if not isinstance(data, list):
            if isinstance(data, pd.DataFrame):
                data = data.values.tolist()
            else:
                raise TypeError("data must be a list or pandas DataFrame")

        url = self.links["upload"]

        content = {"data": data}
        resp = self.session.post(url, json=content)
        return resp.json()

    def update(self, description: Optional[str] = None, tags:  Optional[List[str]] = None, query: Optional[str] = None, category: Optional[str] = None, url: Optional[str] = None):
        """
        Update phenotype attributes
        """
        uri = self.links["self"]

        if tags and not isinstance(tags, list):
            tags = [tags]

        content = {"description": description,
                   "query": query,
                   "tag_list": tags,
                   "category": category,
                   "url": url
                   }

        self.session.patch(uri, json={k: v for k, v in content.items() if v is not None})
        self.refresh()

    def update_query(self, query: str):
        """
        Update the phenotype with a new query that defines this phenotype 
        """
        self.update(query = query)

    def update_description(self, description: str):
        """
        Update the phenotype with a new description
        """
        self.update(description = description)

    def set_tags(self, tags: List[str]):
        """
        Set the tag list for this phenotype, overriding all previous tags
        """
        self.update(tags = tags)

    def get_tags(self):
        """
        Retrieve all tags for this phenotype
        """
        return self.tag_list

    def add_tag(self, tag: str):
        """
        Add a new tag to this phenotype.

        :raises: `PhenotypeError` if the tag is already set on this phenotype
        """
        tags = set(self.tag_list)
        if tag in tags:
            raise PhenotypeError(f"Tag {tag} already exists on this phenotype")

        tags.add(tag)
        self.update(tags = list(tags))

    def delete_tag(self, tag: str):
        """
        Delete a tag from the phenotype.

        :raises: `PhenotypeError` if the tag does not exist
        """
        tags = set(self.tag_list)
        try:
            tags.remove(tag)
        except KeyError:
            raise PhenotypeError(f"Tag {tag} does not exist on this phenotype")

        self.update(tags = list(tags))
    
    def get_info(self):
        """
        Retrieve info for this phenotype
        """
        return self.data

    def get_data(self, label: Optional[str] = None):
        """
        Retrieve a phenotype data from the server.
        """
        matrix = PhenotypeMatrix(self.session, project_name = self.data["project_key"])
        matrix.add_phenotype(name = self.data["name"], label = label)
        self.df = matrix.get_data()
        return self.df

    def display(self, title=None):
        """
        Display phenotype

        :title: Phenotype plot title. Default: Phenotype name.
        """
        if not title:
            title = self.data['name']

        if self.df is None:
            self.get_data()
        
        switcher = {
            "SET": self._plot_set,
            "QT": self._plot_qt,
            "CATEGORY": self._plot_categorical
        }

        init_notebook_mode()

        layout = {
            'title': title,
            'hovermode': False, 
            'showlegend': False,
            'width': 500,
            'height': 400
        }

        fig = switcher.get(self.data.get("result_type"), "Nothing")(layout=layout)
        iplot(fig)

    def _plot_qt(self, **kwargs):
        """
        Plot QT phenotype
        """
        grp_col = self.df.columns[1]
        fig = go.Figure([go.Histogram(x=self.df[grp_col])],
                   **kwargs)
        fig.update_layout(yaxis={'title': "Count"})
        return fig

    def _plot_categorical(self, **kwargs):
        """
        Plot CATEGORICAL phenotype
        """
        grp_col = self.df.columns[1]
        grp_df = self.df.groupby(grp_col).count()
        grp_df = grp_df.reset_index()
 
        fig = go.Figure([go.Bar(x=[str(x) for x in grp_df[grp_col]], y = grp_df['pn'])],
                   **kwargs)
        fig.update_layout(xaxis={'title': "Category"},
                          yaxis={'title': "Count"})
        return fig

    def _plot_set(self, **kwargs):
        """
        Plot SET phenotype
        """
        fig = go.Figure([go.Pie(labels=["Count"], values=[len(self.df.index)])], **kwargs)
        fig.update_traces(textinfo='value')
        return fig



"""
Base Service
------------------
Common service functionality

"""
import os
from urllib.parse import urljoin
from typing import Optional, Dict
import logging
from ..session import ServiceSession
from ..exceptions import InvalidProfile, ServerError

log = logging.getLogger(__file__)


class BaseService:
    """
    Common functionality between all services
    """

    def __init__(self, client, service_path, *args, **kwargs):
        self.client = client

        service_path = os.environ.get("NEXTCODE_SERVICE_PATH", service_path)
        root_url = client.profile.root_url
        if not root_url:
            raise InvalidProfile("Profile is not configured")

        # ! explicit temporary hack
        if os.environ.get("SERVICE_IN_ROOT"):
            service_path = "/"

        self.service_path = service_path
        self.base_url = urljoin(root_url, service_path)
        api_key = client.profile.api_key
        if client.profile.skip_auth:
            api_key = None
        self.session = ServiceSession(self.base_url, api_key)

    def __repr__(self):
        return f"<Service {self.service_name} {self.version} | {self.base_url}>"

    def healthy(self) -> bool:
        """
        Is the service healthy?
        """
        try:
            self.session.get(self.session.endpoints["health"])
            return True
        except ServerError:
            return False

    def status(self, force: Optional[bool] = False) -> Dict:
        """
        Service information from the root endpoint
        """
        if force:
            self.session.fetch_root_info()
        ret = {
            k: v
            for k, v in self.session.root_info.items()
            if k in ("build_info", "app_info", "service_name")
        }
        # include the root endpoint for good measure
        ret["root"] = self.session.root_info.get("endpoints", {}).get("root")
        return ret

    @property
    def service_name(self) -> str:
        return self.session.root_info["service_name"]

    @property
    def version(self) -> str:
        return self.session.root_info["build_info"]["version"]

    @property
    def build_info(self) -> str:
        return self.session.root_info["build_info"]

    @property
    def app_info(self) -> Dict:
        return self.session.root_info.get("app_info")

    @property
    def current_user(self) -> Dict:
        """
        User JWT decoded by the service
        """
        return self.session.root_info.get("current_user") or {}

    @property
    def endpoints(self) -> Dict:
        """
        Dictionary of name: endpoint url for exposed endpoints in the service
        """
        return self.session.endpoints

    def openapi_spec(self) -> Dict:
        """
        Raw openapi spec for the service
        """
        return self.session.get(self.session.url_from_endpoint("documentation")).json()



"""
Service class
------------------
Service object for interfacing with the Workflow Service API.

This class instance is used to communicate with a RESTFul service. The `post_job` method
creates a new job on the server and `find_job` and `get_jobs` allow you to inspect running
and past workflow jobs.

Note: This service has not been fully integrated into the SDK and access is still quite raw. 

"""
import time
import os

from typing import Optional, List, Union, Dict
from ...services import BaseService
from ...client import Client
from ...exceptions import NotFound
from .job import WorkflowJob
from .exceptions import JobError
from ...packagelocal import package_and_upload

import logging

SERVICE_PATH = "workflow"

log = logging.getLogger(__name__)


class Service(BaseService):
    """
    A connection to the workflow service API server
    """

    def __init__(self, client: Client, *args, **kwargs) -> None:
        super(Service, self).__init__(client, SERVICE_PATH, *args, **kwargs)

    def get_pipelines(self) -> List:
        """
        Returns the pipelines available on the current server

        Refer to the API documentation for the Workflow service to see formatting of data.

        :return: List of pipelines
        """
        resp = self.session.get(self.session.url_from_endpoint("pipelines"))
        pipelines = resp.json()["pipelines"]
        return pipelines

    def get_projects(self) -> List:
        """
        Returns the projects that have been created on the current server

        Refer to the API documentation for the Workflow service to see formatting of data.

        :return: List of projects
        """
        resp = self.session.get(self.session.url_from_endpoint("projects"))
        projects = resp.json()["projects"]
        return projects

    def find_job(self, job_id: Union[int, str]) -> WorkflowJob:
        """
        Return a job proxy object
        """
        jobs_endpoint = self.session.url_from_endpoint("jobs")

        data: Dict = {"limit": 1}
        if job_id == "latest":
            data["user_name"] = self.current_user.get("email")
        else:
            try:
                data["job_id"] = int(job_id)
            except ValueError:
                raise NotFound(
                    "job_id must be an integer or 'latest', not '%s'" % job_id
                )
        resp = self.session.get(jobs_endpoint, json=data)
        jobs = resp.json()["jobs"]
        if not jobs:
            raise NotFound("Job not found")

        job = jobs[0]
        return WorkflowJob(self.session, job["job_id"], job)

    def get_jobs(
        self,
        user_name: Optional[str] = None,
        status: Optional[str] = None,
        project: Optional[str] = None,
        pipeline: Optional[str] = None,
        state: Optional[str] = None,
        context: Optional[str] = None,
        limit: Optional[int] = 50,
    ) -> List[WorkflowJob]:
        """
        Get a list of jobs satisfying the supplied criteria

        :param user_name: The user who created the job
        :param status: Current status of jobs
        :param project: Filter by project
        :param pipeline: Filter by pipeline name
        :param state: Filter by state, each state encapsulates several statuses (running, finished)
        :param context: Filter by context string
        :param limit: Maximum number of jobs to return
        """
        data: Dict = {"limit": limit}
        if user_name:
            data["user_name"] = user_name
        if status:
            data["status"] = status
        if project:
            data["project_name"] = project
        if pipeline:
            data["pipeline_name"] = pipeline
        if state:
            data["state"] = state
        if context:
            data["context"] = context
        st = time.time()
        resp = self.session.get(self.session.url_from_endpoint("jobs"), json=data)
        jobs = resp.json()["jobs"]
        log.info("Retrieved %s jobs in %.2f sec", len(jobs), time.time() - st)
        ret = []
        for job in jobs:
            ret.append(WorkflowJob(self.session, job["job_id"], job))
        return ret

    def run_local(
        self,
        path,
        project_name: str = None,
        params: Optional[Dict] = None,
        profile: Optional[str] = None,
        trace: bool = False,
        details: Optional[Dict] = None,
        description: Optional[str] = None,
    ):
        """
        Run a workflow job from a local nextflow folder

        This will package the passed-in folder and upload to a specific bucket that is exported from the workflow
        service on the cluster (see 'scratch-bucket' from the svc.status() method)
        AWS access keys must be set up correctly to upload to this bucket using typical aws/boto environment and
        credential management.

        :param path: Path (relative or absolute) to a folder containing a nextflow script
        :param project_name: Name of the project to run this job on, defaults to GOR_API_PROJECT in environment
        :param params: Dictionary of parameters to forward to the job
        :param profile: Nextflow profile name to use.
        :param trace: Instruct the job to run nextflow trace flags for debugging.
        :param details: Dictionary containing the initial values for 'details' in the job
        :param description: Human readable description of the job
        """

        build_context = package_and_upload(self, "local_workflow", path)
        return self.post_job(
            None,
            project_name,
            build_source="url",
            build_context=build_context,
            profile=profile,
            trace=trace,
            params=params,
            details=details,
            description=description,
        )

    def post_job(
        self,
        pipeline_name: Optional[str] = None,
        project_name: str = None,
        params: Optional[Dict] = None,
        script: Optional[str] = None,
        revision: Optional[str] = None,
        build_source: Optional[str] = None,
        build_context: Optional[str] = None,
        profile: Optional[str] = None,
        trace: bool = False,
        details: Optional[Dict] = None,
        description: Optional[str] = None,
        executor_memory_mb: Optional[int] = None,
        context: Optional[str] = None,
        storage_type: Optional[str] = None,
        dedicated_storage_size: Optional[str] = None,
        keep_resources: Optional[bool] = None,
        mount_jobs: Optional[List[int]] = None,
        credentials: Optional[Dict] = None,
        scheduler_name: Optional[str] = None,
        features: Optional[Dict] = None
    ):
        """
        Run a workflow job

        This is a low-level implementation on top of the workflow service's POST /jobs endpoint.

        :param pipeline_name: Name of the pipeline to run
        :param project_name: Name of the project to run this job on, defaults to GOR_API_PROJECT in environment
        :param params: Dictionary of parameters to forward to the job
        :param script: Git repository url to run (only when server is in development mode)
        :param revision: Git revision or tag (only when server is in development mode)
        :param build_source: Source type of the nextflow build (builtin, git, url)
        :param build_context: Context for build_source, depends on the type.
        :param profile: Nextflow profile name to use.
        :param trace: Instruct the job to run nextflow trace flags for debugging.
        :param details: Dictionary containing the initial values for 'details' in the job
        :param description: Human readable description of the job
        :param executor_memory_mb: Override the memory limit of the nextflow executor
        :param context: Optional string to allow querying for custom information
        :param storage_type: Optional string specifying the storage option to use for the pipeline
        :param dedicated_storage_size Optional string specifying the size of the dedicated volume if using
               storage_type=dedicated
        :param keep_resources: Optional boolean specifying if resources created by workflow service for the Job should
               be exempt from being cleanup up by the workflow resource cleanup process.
        :param mount_jobs: Optional list of job-ids whose work folder will be mounted into this job under
               the folder /job/<job-id>.
        :param credentials: Optional dict containing credentials to forward to workflow-service
        example credentials:  {
            'download': {
                'aws_access_key_id': 'ASIA...',
                'aws_secret_access_key': 'k6q5...'
            },
            'upload': {
                'aws_access_key_id': 'ASIA...',
                'aws_secret_access_key': 'k6q5...'
            }
        }
        :param scheduler_name: Optional string specifying the scheduler to use. If not specified, the default scheduler
               will be used
        :param features: Extra parameters to pass to workflow-service to turn features on/off
        """
        if not project_name:
            project_name = os.environ.get("GOR_API_PROJECT")
        if not project_name:
            raise JobError(
                "No project specified and GOR_API_PROJECT not set in environment"
            )
        log.debug(
            "post_job called with pipeline_name=%s, project_name=%s, params=%s, script=%s, revision=%s, "
            "build_source=%s, build_context=%s, profile=%s, description=%s, executor_memory_mb=%s, context=%s, "
            "storage_type=%s, dedicated_storage_size=%s, mount_jobs=%s scheduler_name=%s",
            pipeline_name,
            project_name,
            params,
            script,
            revision,
            build_source,
            build_context,
            profile,
            description,
            executor_memory_mb,
            context,
            storage_type,
            dedicated_storage_size,
            mount_jobs,
            scheduler_name,
        )
        data = {
            "pipeline_name": pipeline_name,
            "project_name": project_name,
            "parameters": params,
            "revision": revision or None,
            "script": script,
            "build_context": build_context,
            "profile": profile,
            "details": details,
            "description": description,
            "context": context,
            "features": features
        }
        if storage_type:
            data["storage_type"] = storage_type
        if dedicated_storage_size:
            data["dedicated_storage_size"] = dedicated_storage_size
        if build_source:
            data["build_source"] = build_source
        if executor_memory_mb:
            data["executor_memory_mb"] = executor_memory_mb
        if trace:
            data["env"] = {"NXF_DEBUG": "3", "NXF_TRACE": "nextflow"}  # type: ignore
        if credentials:
            data["credentials"] = credentials
        if keep_resources:
            data["keep_resources"] = keep_resources
        if mount_jobs:
            data["mount_jobs"] = mount_jobs
        if scheduler_name:
            data["scheduler_name"] = scheduler_name

        endpoint = self.session.url_from_endpoint("jobs")

        resp = self.session.post(endpoint, json=data)
        job = resp.json()
        return WorkflowJob(self.session, job["job_id"], job=job)



"""
Jupyter Extensions
------------------

Bootstrapping for Jupyter Notebook magic syntax, `%gor` and `%%gor`.

Use `%env LOG_QUERY=1` in Jupyter to see details.
"""
import argparse
import getopt
import hashlib
import threading
import time
import logging
import re
import sys
import os
import datetime
import _thread

import nextcode

from ...exceptions import ServerError, InvalidToken
from .exceptions import MissingRelations, QueryError
from ...utils import jupyter_available, strtobool
from typing import Dict, List, Optional, Union

log = logging.getLogger(__name__)


if jupyter_available():
    """"""
    import pandas as pd
    from IPython.core.magic import (
        Magics,
        magics_class,
        line_magic,
        line_cell_magic,
    )  # type: ignore
    from ipywidgets import IntProgress
    from IPython.display import display
    from tqdm.auto import tqdm, trange

else:

    Magics = object

    def magics_class(cls):
        return cls

    def line_cell_magic(func):
        return func

    def line_magic(func):
        return func


def print_details(txt):
    """
    Print string, but only if LOG_QUERY is set
    """
    if os.environ.get("LOG_QUERY"):
        print(txt, flush=True)


def print_error(txt):
    """
    Print string to stderr
    """
    print(txt, file=sys.stderr, flush=True)


def sizeof_fmt(num, suffix="B"):
    """
    Pretty-printed number format for size.
    """
    import math

    if not num:
        return "-"
    magnitude = int(math.floor(math.log(num, 1024)))
    val = num / math.pow(1024, magnitude)
    if magnitude > 7:
        return "{:.1f}{}{}".format(val, "Yi", suffix)
    return "{:3.1f}{}{}".format(
        val, ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"][magnitude], suffix
    )


def get_service():
    """
    Helper method to get a query service instance
    """
    svc = nextcode.get_service("query")
    return svc


def get_queryserver():
    """
    Helper method to get a query server instance
    """
    svc = nextcode.get_service("queryserver")
    return svc


@magics_class
class GorMagics(Magics):
    """
    The basic 'ipython magic' extension that loads up the sdk as a `%gor` plugin in Jupyter notebook.
    """

    @line_cell_magic
    def gor(self, line, cell=None):
        """
        Execute a gor statement and return the results, supports virtual relations.

        See Jupyter notebook for examples and details.
        """
        try:
            gor_string = line
            return_var = None
            persist = None
            download_filename = None

            # if there is a << on the first line we assume we have an assignment or a persist
            if "<<" in gor_string:
                parts = gor_string.split("<<")
                var = parts[0].strip()
                gor_string = parts[1]

                # if the variable looks like it might be a filename, persist it, otherwise assign to a local var
                if var.startswith("file:"):
                    download_filename = var[5:]
                elif "/" in var or "." in var:
                    persist = var
                else:
                    return_var = var

            # Handle options (use only long options so it is less likely to conflict with gor)
            parser = argparse.ArgumentParser(add_help=False)
            parser.add_argument('--gzip', type=strtobool, nargs='?', const=True, default=True)
            parser.add_argument('--queryservice', type=strtobool, nargs='?', const=True, default=False)
            options, args = parser.parse_known_args(gor_string.split())
            gor_string = ' '.join(args)

            # If this is a multiline statement.
            if cell:
                gor_string += "\n" + cell

            gor_string = self.replace_vars(gor_string)

            if options.queryservice or (persist is not None or download_filename is not None):
                return self.__call_query_service(gor_string, persist, download_filename, return_var)
            else:
                return self.__call_query_server(gor_string, options.gzip, return_var)

        except Exception:
            self.handle_exception()

    @line_magic
    def gorls(self, line):
        """
        List out the contents of the selected folder. Example `%gorls .`
        """
        svc = get_service()
        parts = line.replace("  ", " ").split(" ")
        path = parts[0] or "."
        grep = None
        if len(parts) >= 2:
            grep = parts[1]
        gor_string = (
            f"""nor {path} | SELECT Filename,isDir,FileSize | SORT -c fileName"""
        )
        if grep:
            gor_string += f" | GREP {grep}"
        qry = svc.execute(gor_string)
        if qry.failed:
            print("No results found")
            return None

        results = qry.get_results()
        rows = results.get("data", [])
        if not rows:
            print("No results found")

        ret = []
        for row in rows:
            txt = row[0]
            if row[1] == "true":
                txt += "/"
            else:
                txt += " ({sz})".format(sz=sizeof_fmt(int(row[2])))
            ret.append(txt)
        print("\n".join(ret))
        return None

    @line_magic
    def gorfind(self, line):
        """
        Find a file within the project tree. Example `%gorfind pns.txt`
        """
        svc = get_service()
        string = line
        # okay, you crazy fool. Now we run a bunch of queries to crawl through the entire directory
        # structure of the project to find all folders. Then we run a nor command on each of these
        # folders to grep for the string
        gor_string = f"""nor . -r -d 3 | SELECT Filepath | GREP {string}"""
        qry = svc.execute(gor_string)
        if qry.failed:
            print("No results found")
            return None
        results = qry.get_results()
        folders = results.get("data", [])
        if not folders:
            print("No results found")
        for f in folders:
            ff = f[0].split(svc.project)[-1]
            print(ff)

    def handle_exception(self):
        """
        Print out any exception on the stack.
        """
        exception = sys.exc_info()
        print_error(str(exception[1]))

    def replace_vars(self, string):
        """
        Handle variable substitution in a gor string to interact with local state.
        """
        replacement_vars = re.findall("\\$([^0-9][a-zA-Z0-9_]*)?", string)
        ret = string
        user_ns = self.shell.user_ns
        for var_name in replacement_vars:
            if not var_name:
                continue
            if var_name not in user_ns.keys():
                print("Variable '%s' not found in notebook" % var_name)
                continue
            var = str(user_ns[var_name])
            ret = ret.replace(f"${var_name}", var)
        return ret

    def load_relations(self, relation_names):
        """
        Find relations in local iPython state for inclusion in a remote gor query.
        """
        relations = []
        user_ns = self.shell.user_ns
        print_details(
            "Loading relations {} from local state".format(", ".join(relation_names))
        )
        for name in relation_names:
            var_name = name.replace("[", "").replace("]", "").split(":")[-1]
            if var_name not in user_ns.keys():
                raise Exception("Variable '%s' not found" % var_name) from None
            var = user_ns[var_name]
            if not isinstance(var, pd.DataFrame):
                raise Exception(
                    "%s must be a pandas DataFrame object, not %s"
                    % (var_name, type(var))
                ) from None
            md5 = hashlib.md5()
            md5.update(str(id(user_ns[var_name])).encode())
            data = var.to_csv(index=False, sep="\t")
            relations.append(
                {
                    "name": name,
                    "fingerprint": md5.hexdigest(),
                    "extension": ".tsv",
                    "data": data,
                }
            )
        return relations

    def __call_query_service(self, gor_string, persist, download_filename, return_var):
        try:
            qry = None
            st = time.time()
            svc = get_service()
            try:
                qry = svc.execute(gor_string, nowait=True, persist=persist)
            except MissingRelations as ex:
                relations = self.load_relations(ex.relations)
                qry = svc.execute(
                    gor_string, relations=relations, nowait=True, persist=persist
                )
            poll_time = 1.0
            while qry.running is True:
                t = time.time()
                while time.time() < t + poll_time:
                    diff = int(time.time() - st)
                    diff = str(datetime.timedelta(seconds=diff))
                    sys.stdout.write(f"Query has been running for {diff}...\r")
                    sys.stdout.flush()
                    time.sleep(1.0)
                poll_time = min(poll_time + 1.0, 10.0)

            if qry.status != "DONE":
                try:
                    print_error(
                        "Query {} failed with error:\n{}".format(
                            qry.query_id, qry.status_message
                        )
                    )
                except TypeError:
                    raise Exception(
                        "Query {} has unexpected status {}".format(
                            qry.query_id, qry.status
                        )
                    )
                return None
            num_rows = qry.line_count or 0
            query_time = time.time() - st
            print("Query {} ran in {:.2f} sec".format(qry.query_id, query_time))
            
            if persist:
                return None

            if download_filename:
                from ipywidgets import IntProgress
                from IPython.display import display
                from tqdm.auto import tqdm, trange

                line_count = qry.stats["line_count"]
                f = tqdm(total=line_count, desc=f"Downloading")

                def callback(num_lines):
                    f.update(num_lines)

                ret = qry.download_results(download_filename, callback=callback)
                f.close()
                print(f"Results have been downloaded to {ret}")
                return None
            else:
                MAX_ROWS = 1000000
                if num_rows > MAX_ROWS:
                    print_error(
                        "Query {} returned {} rows but magic commands are capped at {} rows.".format(
                            qry.query_id, qry.line_count, MAX_ROWS
                        )
                    )
                ret = qry.dataframe(limit=MAX_ROWS)

                print("Query {} fetched {:,} rows in {:.2f} sec".format(qry.query_id, num_rows, time.time() - st - query_time))

                if return_var:
                    self.shell.user_ns[return_var] = ret
                    return None
                else:
                    return ret
        except KeyboardInterrupt:
            if qry:
                try:
                    qry.cancel()
                except QueryError:
                    pass
            print_error("Query has been cancelled")
            return None

    def __call_query_server_execute_and_get_dataframe(self, gor_string, gzip, relations, st):
        svc = get_queryserver()
        result = svc.execute(gor_string, gzip=gzip, relations=relations)
        query_time = time.time() - st
        print("Query ran in {:.2f} sec".format(time.time() - st,))

        ret = result.dataframe()

        end_time =  time.time()
        print("Query fetched {:,} rows in {:.2f} sec (total time {:.2f} sec)".format(
            result.num_lines, end_time - st - query_time, end_time - st))

        return ret

    def __call_query_server(self, gor_string, gzip, return_var):
        try:
            result = None
            st = time.time()
            t = updateStatusThread()
            t.start()
            relations = []
            try:
                ret = self.__call_query_server_execute_and_get_dataframe(gor_string, gzip, relations, st)
            except MissingRelations as ex:
                relations = self.load_relations(ex.relations)
                ret = self.__call_query_server_execute_and_get_dataframe(gor_string, gzip, relations, st)

            if return_var:
                self.shell.user_ns[return_var] = ret
                return None
            else:
                return ret

        except KeyboardInterrupt:
            if result:
                try:
                    result.cancel()
                except QueryError:
                    pass
            print_error("Query has been cancelled")
            return None
        finally:
            if t:
                t.stop()
                t.join()


class updateStatusThread (threading.Thread):
    def __init__(self,):
        threading.Thread.__init__(self)
        self.stop_flag = False

    def run(self):
        start_time = time.time()
        last_update_time = start_time
        poll_time = 3.0  # Init poll time.
        while True:
            current_time = time.time()
            if current_time > last_update_time + poll_time:
                last_update_time = current_time
                diff = int(current_time - start_time)
                diff = str(datetime.timedelta(seconds=diff))
                sys.stdout.write(f"Working for {diff}...\r")
                sys.stdout.flush()
                poll_time = 1.0 # Working poll time.

            if self.stop_flag:
                sys.stdout.write(f"                                             \r")
                sys.stdout.flush()
                break
            time.sleep(0.1)

    def stop(self):
        self.stop_flag = True



        # In order to actually use these magics, you must register them with a
# running IPython.
def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    # You can register the class itself without instantiating it.  IPython will
    # call the default constructor on it.
    ipython.register_magics(GorMagics)

    try:
        svc = get_service()
    except InvalidToken:
        print_error("Invalid GOR_API_KEY")
        return
    except ServerError as ex:
        print_error(str(ex))
        return
    status = svc.status(force=True)
    print(
        " Gor magic extension has been loaded. You can now use '%gor' and '%%gor' in this notebook"
    )
    build_info = status["build_info"]
    try:
        gor_version = build_info["query_service"]["gor_version"]
    except KeyError:  # backwards compability
        gor_version = build_info["gor_services_version"]

    print(" * Python SDK Version: {}".format(nextcode.__version__))
    print(" * Query API Version: {}".format(build_info["version"]))
    print(" * GOR Version: {}".format(gor_version))
    print(" * Root Endpoint: {}".format(status["root"]))
    print(" * Current User: {}".format(svc.current_user.get("email")))
    print(" * Current Project: {} (%env GOR_API_PROJECT=xxx)".format(svc.project))


class QueryBuilder:

    defs: Dict[str, str] = {}
    creates: Dict[str, str] = {}

    def __init__(self):
        self.defs = {}
        self.creates = {}

    def render(self, stmt):
        string = ""
        for k, v in self.defs.items():
            v = str(v)
            if not v.endswith(";"):
                v += ";"
            string += "def {} = {}\n".format(k, v)
        string += "\n"
        for k, v in self.creates.items():
            if not v.endswith(";"):
                v += ";"
            string += "create {} = {}\n".format(k, v)
        string += "\n"

        string += stmt
        return string

    def execute(self, stmt, **kw):
        svc = get_service()
        query = self.render(stmt)
        return svc.execute(query, **kw)



"""
Service class
------------------
Service object for interfacing with the GOR Query Server.

This class instance is used to communicate with a RESTFul service. Use the `execute`
methods to get started.

"""
import json
import logging
import os
from typing import Dict, Tuple, Sequence, List, Optional, Union, Any

from .result import Result
from .result import throw_error_from_line
from ...services import BaseService
from ...exceptions import ServerError
from ...client import Client
from ..query.exceptions import QueryError

from ..query.utils import extract_virtual_relations
import nextcode

SERVICE_PATH = "queryserver"

log = logging.getLogger(__name__)


class Service(BaseService):
    """
    A connection to the query server
    """

    def __init__(self, client: Client, *args, **kwargs) -> None:
        super(Service, self).__init__(client, SERVICE_PATH, *args, **kwargs)
        self.metadata = {"client": f"nextcode-python-sdk/{nextcode.__version__}"}
        self.project = (
            kwargs.get("project")
            or os.environ.get("GOR_API_PROJECT")
            or client.profile.project
        )

    def _check_project(self):
        """
        Raise an exception if there is no project specified

        :raises: QueryError
        """
        if not self.project:
            raise QueryError("No project specified")

    def set_project(self, project: str, persist: bool = True) -> None:
        """
        Set the current project for all subsequent queries.

        :project: The project name
        :persist: Persist the project name into the current profile

        """
        self.project = project
        if persist:
            self.client.profile.project = self.project

    def execute(
        self,
        query: str,
        relations: Optional[List[Dict]] = None,
        job_type: Optional[str] = None,
        gzip: Optional[bool] = False,
        **kw,
    ):  # -> SynchronousQuery:
        """
        Execute a gor statement on the server synchronously.

        :param query: gor query string
        :param relations: virtual relations to include with the query
        :param job_type: Optional job type for routing purposes
        :param gzip:Optional if result should be gzipped.

        :raises: :exc:`~exceptions.ServerError`, :exc:`~exceptions.MissingRelations`

        Optional keyword arguments in the form `name=data` (where data is a tsv string) are converted into virtual relations if relations
        is not explicitly passed in and added to the relations set by the relations parameter. If the relations parameter is set it
        should be a list of dictionaries with {"name": "relation-name", "data": "<tsv string>"}
        """
        self._check_project()

        payload_relations = extract_virtual_relations(kw, relations)

        payload: Dict[str, Optional[Union[int, str, List[Any], Dict]]] = {
            "project": self.project,
            "query": query,
            "user": "python-sdk",
            "virtualRelations": payload_relations,
            "routingKey": job_type or "default",
            "useGzip": str(gzip).lower(),
            "sendTerm": True,
            "sendProgress": True,
            "sendAlive": True
        }

        url = self.session.url_from_endpoint("query")

        try:
            resp = self.session.post(url,
                                     json=payload,
                                     stream=True,
                                     headers={"Content-Type": "application/json",
                                              "Accept-Encoding": "chunked",
                                              "Accept": "application/octet-stream"})
            resp.raise_for_status()
        except ServerError as ex:
            throw_error_from_line(ex.message)

        return Result(resp, gzip)



"""
keycloak
~~~~~~~~~~
Keycloak login helpers and management features
"""

import os
from posixpath import join as urljoin
import requests
from requests import codes
import logging
import webbrowser
from typing import Optional, Dict, List, Any

from .exceptions import ServerError, AuthServerError
from .utils import host_from_url

log = logging.getLogger(__name__)

DEFAULT_REALM = "wuxinextcode.com"
DEFAULT_CLIENT_ID = "api-key-client"


def login_keycloak_user(
    root_url: str,
    user_name: str,
    password: str,
    realm: str = DEFAULT_REALM,
    client_id: str = DEFAULT_CLIENT_ID,
) -> str:
    """
    :param root_url: The root url of the service. e.g. www.myhost.com
    :param user_name: Local keycloak username
    :param password: Password of the user
    :param realm: Realm to authenticate against. Default: wuxinextcode.com
    :param client_id: Client to log in as. Default: api-key-client

    :raises: `AuthServerError`

    Logs into the keycloak server using a username and password. Returns the api key.
    """
    # try logging in with the new user
    root_url = host_from_url(root_url)
    url = urljoin(root_url, "auth", "realms", realm, "protocol/openid-connect/token")
    body = {
        "grant_type": "password",
        "client_id": client_id,
        "password": password,
        "username": user_name,
        "scope": "offline_access",
    }
    url = url.replace("admin/", "")
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    log.debug("Calling POST %s with headers %s and body %s", url, headers, body)
    resp = requests.post(url, headers=headers, data=body)
    try:
        resp.raise_for_status()
        return resp.json()["refresh_token"]
    except Exception as ex:
        raise AuthServerError(f"User {user_name} was unable to log in: {ex}")


def auth_popup(root_url: str) -> None:
    """
    Opens a web browser on the login page for the specified server

    The user can log in using OAuth2 flow and aquire an API Key
    """
    root_url = host_from_url(root_url)
    url = urljoin(root_url, "api-key-service")
    webbrowser.open(url)


def _get_admin_keycloak_session(auth_server, username, password):
    log.info("Managing users on keycloak server %s...", auth_server)

    # log the admin into the master realm
    url = urljoin(auth_server, "realms/master/protocol/openid-connect/token")
    headers_form = {"Content-Type": "application/x-www-form-urlencoded"}
    resp = requests.post(
        url,
        data=f"username={username}&password={password}&grant_type=password&client_id=admin-cli",
        headers=headers_form,
    )
    if resp.status_code != 200:
        desc = resp.json()["error_description"]
        raise AuthServerError(
            f"Could not authenticate {username} user against {url}: {desc}"
        )
    access_token = resp.json()["access_token"]
    headers = {
        "Authorization": "Bearer %s" % access_token,
        "Content-Type": "application/json",
    }
    session = requests.Session()
    session.headers = headers
    return session


class KeycloakSession:
    """
    A session object on a keycloak server intended to provide high-level interfaces
    for managing users and roles on the server.

    The user that is passed into the constructor is assumed to be a keycloak admin.

    :param root_url: The URL of the instance where keycloak is running. e.g. www.myhost.com
    :param username: Username of the keycloak admin. Default: admin
    :param passwrod: Admin password
    :param realm: Keycloak realm to manage. Default: wuxinextcode.com
    :param client_id: Client to log admin into. Default: api-key-client
    """

    def __init__(
        self,
        root_url: str,
        username: str = "admin",
        password: Optional[str] = None,
        realm: str = DEFAULT_REALM,
        client_id: str = DEFAULT_CLIENT_ID,
    ):
        password = password or os.environ.get("KEYCLOAK_PASSWORD")
        if not password:
            raise AuthServerError(f"User {username} needs an admin password")
        self.root_url = host_from_url(root_url)
        self.realm = realm
        self.auth_server = self._get_auth_server()
        self.session = _get_admin_keycloak_session(self.auth_server, username, password)
        self.realm_url = urljoin(self.auth_server, "admin/realms", realm)
        self.master_url = urljoin(self.auth_server, "admin/realms", "master")
        self.client_id = client_id

    def _get_auth_server(self):
        auth_server = urljoin(self.root_url, "auth")
        r = requests.get(auth_server)

        # test if realm is available
        realm_url = urljoin(auth_server, "realms", self.realm)
        try:
            resp = requests.get(realm_url)
        except requests.exceptions.ConnectionError:
            raise ServerError(f"Keycloak server {realm_url} is not reachable")
        if resp.status_code == requests.codes.not_found:
            raise AuthServerError(
                f"Realm '{self.realm}' was not found on keycloak server {auth_server}"
            )

        return auth_server

    def get_user(self, user_name: str) -> Dict:
        users_url = urljoin(self.realm_url, f"users?username={user_name}")
        resp = self.session.get(users_url)
        resp.raise_for_status()
        users = [u for u in resp.json() if u.get("username") == user_name]
        if not users:
            raise AuthServerError(f"User '{user_name}' not found.")
        return users[0]

    def get_users(self) -> List:
        users_url = urljoin(self.realm_url, "users")
        resp = self.session.get(users_url)
        resp.raise_for_status()
        if not resp.json():
            return []
        users = resp.json()
        return users

    def remove_user(self, user_name: str) -> None:
        user_id = self.get_user(user_name)["id"]

        url = urljoin(self.realm_url, "users", user_id)
        resp = self.session.delete(url)
        resp.raise_for_status()
        log.info("User '%s' has been deleted from realm '%s'", user_name, self.realm)

    def get_user_roles(self, user_name: str) -> List[str]:
        """
        Get the effective realm roles for this user, ignoring client-specific roles
        """
        user_id = self.get_user(user_name)["id"]
        url = urljoin(self.realm_url, f"users/{user_id}/role-mappings/realm/composite")
        resp = self.session.get(url)
        resp.raise_for_status()
        roles = resp.json()
        role_names = [r["name"] for r in roles]
        return role_names

    def create_user(
        self, user_name: str, new_password: str, exist_ok: bool = False
    ) -> None:
        user_id = None
        try:
            user = self.get_user(user_name)
        except AuthServerError:
            pass
        else:
            if not exist_ok:
                raise AuthServerError(f"User '{user_name}' already exists.")
            else:
                user_id = user["id"]

        if not user_id:
            url = urljoin(self.realm_url, "users")
            data = {
                "username": user_name,
                "firstName": "None",
                "lastName": "None",
                "email": user_name,
                "enabled": True,
                "emailVerified": True,
            }
            resp = self.session.post(url, json=data)
            if resp.status_code != codes.created:
                msg = resp.text
                try:
                    msg = resp.json()["errorMessage"]
                except Exception:
                    pass
                raise AuthServerError(msg)
            resp.raise_for_status()
            user_id = self.get_user(user_name)["id"]

        self.set_user_password(user_name, new_password)
        log.info("User '%s' has been created in realm '%s'.", user_name, self.realm)

    def delete_user(self, user_name: str) -> None:
        user_id = self.get_user(user_name)["id"]

        url = urljoin(self.realm_url, "users", str(user_id))
        resp = self.session.delete(url)
        resp.raise_for_status()
        log.info("User '%s' has been deleted from realm '%s'" % (user_name, self.realm))

    def add_role_to_user(
        self, user_name: str, role_name: str, exist_ok: bool = False
    ) -> None:
        user_id = self.get_user(user_name)["id"]
        role_name = role_name.lower()
        user_roles = self.get_user_roles(user_name)
        if role_name in user_roles:
            if exist_ok:
                return
            raise AuthServerError(f"User '{user_name}' already has role '{role_name}'")

        url = urljoin(self.realm_url, f"users/{user_id}/role-mappings/realm/available")
        resp = self.session.get(url)
        resp.raise_for_status()
        available_roles = resp.json()
        role = None
        for r in available_roles:
            if r["name"].lower() == role_name:
                role = r
                break
        if not role:
            raise AuthServerError(
                f"Role '{role_name}' is not available for user '{user_name}'"
            )

        url = urljoin(self.realm_url, f"users/{user_id}/role-mappings/realm")
        resp = self.session.post(url, json=[role])
        resp.raise_for_status()
        log.info(
            "Role '%s' has been added to user '%s' in realm '%s'",
            role_name,
            user_name,
            self.realm,
        )

    def remove_role_from_user(self, user_name: str, role_name: str) -> None:
        user_id = self.get_user(user_name)["id"]
        role_name = role_name.lower()

        url = urljoin(self.realm_url, f"users/{user_id}/role-mappings")
        resp = self.session.get(url)
        resp.raise_for_status()
        roles = resp.json()["realmMappings"]
        role = None
        for r in roles:
            if r["name"].lower() == role_name:
                role = r
                break
        if not role:
            raise AuthServerError(f"User '{user_id}' does not have role '{role_name}'")

        url = urljoin(self.realm_url, "users/%s/role-mappings/realm" % (user_id))
        resp = self.session.delete(url, json=[role])
        resp.raise_for_status()
        log.info(
            "Role '%s' has been removed from user '%s' in realm '%s'",
            role_name,
            user_name,
            self.realm,
        )

    def get_user_client_roles(self, user_name: str, client_name: str) -> List[str]:
        """
        Get the specified client roles for this user
        """
        user_id = self.get_user(user_name)["id"]
        client_id = self.get_client(client_name)["id"]
        url = urljoin(self.realm_url, "users", str(user_id), "role-mappings","clients",client_id)
        resp = self.session.get(url)
        resp.raise_for_status()
        roles = resp.json()
        role_names = [r["name"] for r in roles]
        return role_names

    def add_client_role_to_user(
            self, user_name: str, client_name: str, role_name: str, exist_ok: bool = False
    ) -> None:
        user_id = self.get_user(user_name)["id"]
        user_client_roles = self.get_user_client_roles(user_name,client_name)
        if role_name in user_client_roles:
            if exist_ok:
                return
            raise AuthServerError(f"User '{user_name}' already has role '{role_name}' for client '{client_name}'")

        client_id = self.get_client(client_name)["id"]
        url = urljoin(self.realm_url, f"users/{user_id}/role-mappings/clients/{client_id}/available")
        resp = self.session.get(url)
        resp.raise_for_status()
        available_roles = resp.json()
        role = None
        for r in available_roles:
            if r["name"].lower() == role_name:
                role = r
                break
        if not role:
            raise AuthServerError(
                f"Role '{role_name}' for client '{client_name}' is not available for user '{user_name}'"
            )

        url = urljoin(self.realm_url, f"users/{user_id}/role-mappings/clients",client_id)
        resp = self.session.post(url, json=[role])
        resp.raise_for_status()
        log.info(
            "Role '%s' for client '%s' has been added to user '%s' in realm '%s'",
            role_name,
            client_name,
            user_name,
            self.realm,
        )

    def get_client(self, client_name: str) -> Dict:
        url = urljoin(self.realm_url, f"clients?clientId={client_name}")
        resp = self.session.get(url)
        resp.raise_for_status()
        for c in resp.json():
            if c.get("clientId") == client_name:
                return c
        raise AuthServerError(f"Client '{client_name}' not found.")

    def login_user(self, user_name: str, password: str) -> Optional[str]:
        try:
            return login_keycloak_user(
                self.root_url, user_name, password, self.realm, self.client_id
            )
        except AuthServerError:
            log.exception("Unable to log in")
            return None

    def set_user_password(self, user_name: str, new_password: str) -> None:
        user_id = self.get_user(user_name)["id"]

        url = urljoin(self.realm_url, f"users/{user_id}/reset-password")
        data = {"type": "password", "temporary": False, "value": new_password}
        resp = self.session.put(url, json=data)
        resp.raise_for_status()
        api_key = self.login_user(user_name, new_password)
        if not api_key:
            raise Exception("Changed password but could not log in!")

    def get_available_roles_for_user(self, user_name: str) -> List[str]:
        user_id = self.get_user(user_name)["id"]
        url = urljoin(self.realm_url, f"users/{user_id}/role-mappings/realm/available")
        resp = self.session.get(url)
        resp.raise_for_status()
        available_roles = resp.json()
        return [r["name"] for r in available_roles]

    def create_role(
        self, role_name: str, description: str = "N/A", exist_ok: bool = False
    ) -> None:
        url = urljoin(self.realm_url, "roles")
        data = {"name": role_name, "description": description}
        resp = self.session.post(url, json=data)
        if resp.status_code == codes.conflict:
            if exist_ok:
                return
            raise AuthServerError(f"Role '{role_name}' already exists")
        resp.raise_for_status()

    def get_roles(self) -> Dict[str, Any]:
        url = urljoin(self.realm_url, "roles")
        resp = self.session.get(url)
        resp.raise_for_status()
        roles = resp.json()
        ret = {}
        for r in roles:
            ret[r["name"]] = r
        return ret

    def delete_role(self, role_name: str) -> None:
        role = self.get_roles().get(role_name)
        if not role:
            raise AuthServerError(f"Role '{role_name}' does not exist")
        url = urljoin(self.realm_url, f"roles-by-id/{role['id']}")
        resp = self.session.delete(url)
        resp.raise_for_status()



"""
Pytest plugin
------------------

This is a pytest plugin that is used by workflow CI jobs for automatically testing against a live server.

"""
from nextcode.credentials import generate_credential_struct, creds_to_dict
from nextcode.packagelocal import package_and_upload
from nextcode import get_service

from urllib.parse import urlsplit

import logging
import time
import boto3
import os
import posixpath
import tempfile
import pytest
import sys
import uuid
import pprint


def pytest_addoption(parser):
    group = parser.getgroup("nc-wf")
    group.addoption(
        "--project",
        action="store",
        dest="project",
        help="Sets the name of the project to run the pipeline in",
        default="no_project",
    )
    group.addoption(
        "--runid",
        action="store",
        dest="run_id",
        help="The name/id of the folder the pipeline result_dir will be stored in",
        default=str(uuid.uuid4()),
    )
    group.addoption(
        "--runmode",
        action="store",
        dest="run_mode",
        type=str,
        choices=["repository", "local"],
        help="Specifies if you want to run your local copy of the pipeline, or a version from a remote git repository",
        default="local",
    )
    group.addoption(
        "--revision",
        action="store",
        dest="revision",
        help="Specifies what revision should be run of the pipeline repository",
    )
    group.addoption(
        "--repository",
        action="store",
        dest="repository",
        help="Specifies the git repository containing the pipeline",
    )
    group.addoption(
        "--localdir",
        action="store",
        dest="local_dir",
        help="Specifies what directory contains the pipeline",
        default=".",
    )
    group.addoption(
        "--profile",
        action="store",
        dest="profile",
        help="Specifies the default profile used to run the pipeline",
        default="test",
    )
    group.addoption(
        "--use-last-run",
        action="store_true",
        dest="use_last_run",
        help="If specified the wf plugin will try to use the output from the last run instead of starting a new one",
    )
    group.addoption(
        "--storage_type",
        action="store",
        dest="storage_type",
        help="Specifies the storage type to use for the pipeline run",
        default=None,
    )
    group.addoption(
        "--credentials",
        action="append",
        nargs="+",
        dest="credentials",
        help="Specifies the credentials to forward to workflow-service",
        default=[],
    )
    group.addoption(
        "--scheduler_name",
        action="store",
        dest="scheduler_name",
        help="Specifies the scheduler to use in the pipeline run",
        default=None,
    )
    parser.addini(
        "base_upload_bucket",
        "The S3 location the result_dir parameter of the pipeline will be set to",
        default="s3://nextcode-scratch/workflow-ci",
    )


# Let's suppress the wild logging of these libraries
logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.WARNING)
logging.getLogger("boto3").setLevel(logging.INFO)
logging.getLogger("botocore").setLevel(logging.WARNING)
logging.getLogger("s3transfer").setLevel(logging.INFO)
logging.getLogger("nextcode").setLevel(logging.INFO)
log = logging.getLogger(__name__)

start_time = 0


def pytest_report_header(config, startdir):

    return (
        "Running workflow in project '%s'. This can take some time so going for a coffee is not a bad idea :)"
        % config.option.project
    )


def print_status(txt, newline=True, stamp=True):
    d = time.time() - start_time
    if newline and stamp:
        print("[%.2fs] %s" % (d, txt), flush=True)
    elif newline:
        print(txt, flush=True)
    else:
        print(txt, flush=True, end="")


@pytest.fixture(scope="class", autouse=False)
def run_workflow(request):
    svc = get_service("workflow")
    global start_time
    start_time = time.time()
    class_id = request.cls.__name__
    print_status("WF plugin starting.")
    # Check if we should use the files generated by last run instead of starting a new one.
    last_run_key = "wf_plugin/latest_result_dir/{}".format(class_id)
    use_last_run = request.config.option.use_last_run
    if use_last_run:
        print_status(
            "--use-last-run flag set. Using output files from the previous run."
        )
        prev_dir = request.config.cache.get(last_run_key, None)
        if prev_dir:
            print_status("Previous file dir found: %s." % prev_dir)
            # use the previous directory if it exists
            if os.path.exists(prev_dir):
                request.cls.result_dir = prev_dir
                return
            # previous directory doesn't exists.  Remove the key and continue with the run.
            print_status(
                "WARN: '%s' doesn't exist anymore.  Rerunning workflow." % prev_dir
            )
            request.config.cache.set(last_run_key, None)
        else:
            print_status("Previous file dir not found. Starting a new run.")

    base_result_dir = request.config.getini("base_upload_bucket")
    project_name = request.config.option.project
    run_mode = request.config.option.run_mode
    run_id = request.config.option.run_id
    storage_type = request.config.option.storage_type
    credentials_args = request.config.option.credentials
    scheduler_name = request.config.option.scheduler_name
    # Use the profile specified by the TestCase if it exists. Use the default profile if none is provided.
    profile = getattr(request.cls, "profile", request.config.option.profile)
    revision = None

    if run_mode == "repository":
        build_context = request.config.option.repository
        revision = request.config.option.revision
        build_source = "git"
        if not build_context:
            print("ERROR: Your must specify a --repository when --runmode = repository")
            sys.exit(-1)

    else:
        path = os.path.abspath(os.path.expanduser(request.config.option.local_dir))
        p = package_and_upload(svc, "local_workflow", path)
        build_source = "url"
        build_context = p

    # Add the parameters if the TestCase defines them
    params = getattr(request.cls, "params", {})

    # add result_dir
    result_dir = "{prefix}/{run}/{instance}/".format(
        prefix=base_result_dir, run=run_id, instance=class_id
    )
    params["result_dir"] = result_dir
    print_status(
        "Workflow run parameters:\n\tWorkflow repository: %s:%s\n\tProject: %s\n\tProfile: %s\n\tParameters: %s"
        % (
            build_context,
            revision,
            project_name,
            profile,
            pprint.pformat(params, width=-1),
        )
    )
    cred_map = creds_to_dict([x[0] for x in credentials_args])
    credentials = generate_credential_struct(cred_map)
    job = svc.post_job(
        None,
        project_name,
        params,
        build_context,
        revision,
        build_source=build_source,
        build_context=build_context,
        profile=profile,
        storage_type=storage_type,
        credentials=credentials,
        scheduler_name=scheduler_name,
    )

    last_status = ""
    while job.running:
        time.sleep(5)
        job.refresh()
        if last_status == job.status:
            print_status(".", newline=False)
        else:
            print_status(
                "\nJob %s status = %s" % (job.job_id, job.status), newline=False
            )
        last_status = job.status

    print_status(
        "\nJob %s is no longer running. Status = %s" % (job.job_id, job.status)
    )

    if not job.status == "COMPLETED":
        print_status(job.logs("pod", ""))
        raise Exception("Job did not complete successfully")

    temp_dir = tempfile.mkdtemp()

    _download_all(result_dir, temp_dir)

    request.cls.result_dir = temp_dir
    # Store the latest result dir in cache
    request.config.cache.set(last_run_key, temp_dir)
    print_status("WF plugin run finished.")


def _download_all(s3_path, local_path):
    """Download all files from s3_path into the local folder"""
    s3 = boto3.resource("s3")
    parts = urlsplit(s3_path)
    bucket_name = parts.netloc
    bucket = s3.Bucket(bucket_name)  # pylint: disable=no-member
    print_status(
        "\nDownloading from '{}' to '{}':".format(s3_path, local_path), newline=False
    )
    root_remote_dir = parts.path[1:]

    try:
        os.makedirs(local_path)
    except IOError:
        pass

    for f in bucket.objects.filter(Prefix=root_remote_dir):

        if f.key[-1] == "/":  # skip directories
            continue

        # Get the relative name of the object
        rel_path = posixpath.relpath(f.key, root_remote_dir)

        local_file = os.path.join(local_path, rel_path)
        try:
            os.makedirs(os.path.dirname(local_file))
        except IOError:
            pass
        log.info("Downloading file %s" % f.key)
        print_status(".", newline=False)
        bucket.download_file(f.key, local_file)



"""
Utility methods
------------------

Utility methods used by the query service.
"""

import hashlib
from typing import Dict, Tuple, Sequence, List, Optional, Union, Any

from nextcode.services.query.exceptions import QueryError

DEFAULT_EXTENSION = ".tsv"


def get_fingerprint(contents: str) -> str:
    """
    Generate a fingerprint for the contents of a virtual relation.

    This fingerprint is used by the server for caching purposes.

    :param contents: The full contents of a tsv file
    :returns: md5 sum representing the file contents
    """
    md5 = hashlib.md5()
    md5.update(repr(contents).encode())
    return md5.hexdigest()


def extract_virtual_relations(kw, relations: Optional[List[Dict]] = None):
    """
    Extract virtual relations from input arguments
    :param kw:          optional keyword arguments
    :param relations:   optional relations
    :return:  relations ready for payload
    """
    _relations: List[Dict] = []
    if relations:
        _relations = relations
    else:
        for k, v in kw.items():
            _relations.append({"name": f"[{k}]", "data": v})

    payload_relations: List[Dict] = []
    for r in _relations:
        if "name" not in r or ("data" not in r and "fingerprint" not in r):
            raise QueryError("Virtual relations must have name and eiter data or fingerprint fields")

        name = r["name"]
        data = r["data"] if "data" in r else None

        if data is not None:
            if hasattr(data, "to_csv"):
                data = data.to_csv(index=False, sep="\t")

            if not isinstance(data, str):
                raise QueryError(f"Virtual relation data for {name} must be a string")

            if not data.startswith("#"):
                data = "#" + data

        fingerprint = r.get("fingerprint") or get_fingerprint(data)
        extension = r.get("extension") or DEFAULT_EXTENSION
        payload_relations.append(
            {
                "name": name,
                "fingerprint": fingerprint,
                "extension": extension,
                "data": data,
            }
        )
    return payload_relations



"""
Service class
------------------
Service object for interfacing with the Phenoteke API

"""

import os
import logging
from posixpath import join as urljoin
from typing import Optional, List, Union, Dict

from ...client import Client
from ...services import BaseService

SERVICE_PATH = "phenoteke/api/v1"

log = logging.getLogger(__file__)

class Service(BaseService):
    """
    Phenoteke Service
    """

    links: Dict = {}

    def __init__(self, client: Client, *args, **kwargs) -> None:
        super(Service, self).__init__(client, SERVICE_PATH, *args, **kwargs)

    def map_cids_to_ipns(
            self,
            site_id: str,
            study_id: str,
            cids: List[str],
            subcategory_id: Optional[str] = None,
    ) -> List:
        """
        Get mapping from collaborator ids to ipns

        :param site_id: Site code
        :param study_id: Study code
        :param cids: List of collaborator ids
        :param subcategory_id: Optional sub category within study
        :return: List of {"cid": <cid>, "ipn": <ipn>}
        :raises: ServerError
        """
        uri = urljoin(
            self.session.url_from_endpoint("root"),
            "cids_to_ipns_mapping",
        )
        result = []
        parts = int((len(cids)+4)/5)

        payload = {
            "site_id": site_id,
            "study_id": study_id,
        }
        if subcategory_id:
            payload["subcategory_id"] = subcategory_id

        def list_split(l, parts):
            length = len(l)
            return [ l [i*length // parts: (i+1)*length // parts] for i in range(parts) ]

        for cid_parts in list_split(cids,parts):
            payload["cids"] = cid_parts

            resp = self.session.post(uri, json=payload)
            resp.raise_for_status()
            data = resp.json()
            result.extend(data["data"])

        return result




"""
Analysis catalog
------------------

Analysis catalog is a collection of phenotypes and parameters that are needed to run
the workflow analysis


"""

import json
import datetime
import dateutil
import time
import logging
from posixpath import join as urljoin
from typing import Callable, Union, Optional, Dict, List

from .exceptions import PhenotypeError
from ...exceptions import ServerError
from ...session import ServiceSession

log = logging.getLogger(__name__)


class AnalysisCatalog:
    """
    A local object representing an analysis catalog response from the phenotype
    catalog service.

    Please refer to the API Documentation for the phenotype catalog service.

    In addition to the ones documented here, this object has at least these attributes:

    * name - Analysis Catalog name
    * recipe_name - The name of the recipe to use
    * recipe_parameters - The parameters required to run the recipe
    * analysis_catalog_items - phenotypes/covariates used as input when generating preparation data
    * excluded_pns - the PNs to exclude from the analysis
    * created_at - Timestamp when the analysis catalog was first created
    * updated_at - Timestamp when the analysis catalog was last updated
    * created_by - Username who created the analysis catalog

    """

    def __init__(self, session: ServiceSession, data: Dict):
        self.session = session
        self.data = data

    def __getattr__(self, name):
        try:
            val = self.data[name]
        except KeyError:
            raise AttributeError

        try:
            val = dateutil.parser.parse(val)
        except Exception:
            pass

        return val

    def __repr__(self) -> str:
        return f"<AnalysisCatalog {self.name} in project {self.project_key}>"



"""
Query Object
------------------

The Query class represents a query model from the RESTFul Query API.

There are several helper methods available here but you can view the
raw server response in the `raw` member.

Note that the object is dynamic and some of the attributes available are not
documented here because they are added from the server response.
"""

import logging
import time
from typing import Dict, Tuple, Sequence, List, Optional, Union, Callable
from dateutil.parser import parse
import os

from io import StringIO

from .exceptions import QueryError
from ...utils import jupyter_available

SERVICE_PATH = "/api/query"

RUNNING_STATUSES = ("PENDING", "RUNNING", "CANCELLING")
FAILED_STATUSES = ("CANCELLING", "CANCELLED", "FAILED")
RESULTS_PAGE_SIZE = 1000000

log = logging.getLogger(__name__)


def _log_download_progress(
    num_chunk,
    num_bytes,
    num_lines,
    total_received_lines,
    total_expected_lines,
    start_time,
    callback=None,
):
    """
    Helper for logging out download progress and notifying a listener or delta lines
    """
    mb = num_bytes / 1024 / 1024
    if callback:
        callback(num_lines)
    diff = time.time() - start_time
    if num_chunk is not None and num_chunk % 1000 == 0:
        msg = f"Downloaded {total_received_lines}/{total_expected_lines} lines ({mb:.2f} MB) in {diff:.2f} sec"
        log.info(msg)


class Query:
    """
    A local proxy representing a serverside Gor Query.
    This object is returned from query methods in the 'query' service.

    :param service: query service handle
    :param resp: json response from query api endpoint

    Example usage:

    >>> svc = nextcode.Client(profile_name='test').service('query', project='myproject')
    >>> query = svc.get_queries(limit=1)[0]
    >>> query.status
    'DONE'
    """

    raw: Dict = {}
    query_id = None
    url = None
    duration = None
    query = None
    status = None

    def __init__(self, service, resp: Optional[Dict] = None):
        self.service = service
        self.session = service.session
        if resp:
            self.init_from_resp(resp)

    def init_from_resp(self, resp: Dict):
        """
        Initialize the query object from a server json response.
        If the response is a 'simple query', e.g. from the queries list
        the object will only contain partial information.
        If the caller requests a field that is not included in the local proxy object
        the query will be refreshed from the server.
        """
        # make sure the url is set

        self.url = resp["links"]["self"]
        self.raw = resp
        self.__dict__.update(resp)
        self.submit_date = parse(resp["submit_date"])
        if "stats" in resp:
            stats = resp["stats"]
            self.__dict__.update(resp["stats"])
            if stats.get("end_timestamp"):
                self.duration = stats.get("end_timestamp") - stats.get(
                    "submit_timestamp"
                )

    def __repr__(self):
        return f"<GorQuery {self.query_id or 'NEW'} ({self.status})>"

    def __getattr__(self, name):
        # if we cannot find the attribute we refresh the query from the server
        # just in case we have a partial object
        self.refresh()
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            raise AttributeError()

    @property
    def running(self) -> bool:
        """
        Is the query currently running
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in RUNNING_STATUSES

    @property
    def failed(self) -> bool:
        """
        Is the query in a failed state
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in FAILED_STATUSES

    @property
    def done(self) -> bool:
        """
        Is the query in the DONE state
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status == "DONE"

    @property
    def perspectives(self) -> List[str]:
        """
        Returns a list of perspectives available for this query.

        This only applies to template queries. If this is an ad-hoc query the list will be empty.
        """
        perspective_links = self.raw["links"].get("perspectives") or []
        ret = [p["name"] for p in perspective_links]
        return ret

    def wait(self, max_seconds: Optional[int] = None, poll_period: float = 0.5):
        """
        Wait for the query to complete

        :param max_seconds: raise an exception if the query runs longer than this
        :param poll_period: Number of seconds to wait between polling (max 10 seconds)

        :raises: QueryError
        """
        if not self.running:
            return self
        log.info("Waiting for query %s to complete...", self.query_id)
        start_time = time.time()
        duration = 0.0
        period = poll_period
        is_running = self.running
        while is_running:
            time.sleep(period)
            duration = time.time() - start_time
            is_running = self.running
            if is_running and max_seconds and duration > max_seconds:
                raise QueryError(
                    f"Query {self.query_id} has exceeded wait time {max_seconds}s and we will not wait any longer. It is currently {self.status}."
                )
            period = min(period + 0.5, 10.0)
        if self.status == "DONE":
            log.info(
                "Query %s completed in %.2f sec and generated %s rows",
                self.query_id,
                duration,
                self.line_count,
            )
        else:
            log.info(
                "Query %s has status %s after %.2f sec",
                self.query_id,
                self.status,
                duration,
            )
        return self

    def refresh(self):
        """
        Refresh the local query object from the RESTful service
        """
        if not self.url:
            return
        resp = self.session.get(self.url)
        self.init_from_resp(resp.json())
        return self

    def download_results(self, filename, callback=None):
        """
        Download the entire results for the query to a local file in a single call
        via streaming.
        :param filename: Local filename to save results to
        :raises: QueryError if data is missing.
        """
        filename = os.path.expanduser(filename)
        start_time = time.time()
        try:
            url = self.links["streamresults"]
        except KeyError:
            raise Exception(
                "Server does not support result downloading via streaming"
            ) from None
        num_bytes = 0
        num_lines = 0
        total_expected_lines = 0
        with open(filename, "wb") as f:
            with self.session.get(
                url, headers={"Accept": "text/tab-separated-values"}, stream=True
            ) as r:
                r.raise_for_status()
                total_expected_lines = int(r.headers.get("Line-Count", -1))
                log.info(f"Starting to stream {total_expected_lines} lines...")
                total_received_lines = 0
                for i, chunk in enumerate(r.iter_content(chunk_size=8192)):
                    num_bytes += len(chunk)
                    num_lines = chunk.count(b"\n")
                    total_received_lines += num_lines
                    f.write(chunk)

                    _log_download_progress(
                        i,
                        num_bytes,
                        num_lines,
                        total_received_lines,
                        total_expected_lines,
                        start_time,
                        callback,
                    )

        _log_download_progress(
            0,
            num_bytes,
            num_lines,
            total_received_lines,
            total_expected_lines,
            start_time,
            callback,
        )
        if total_received_lines < total_expected_lines:
            raise QueryError(
                f"Downloaded {total_received_lines} lines but {total_expected_lines} lines expected"
            )
        return filename

    def get_results(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort: Optional[str] = None,
        filt: Optional[str] = None,
        perspective: Optional[str] = None,
        is_json: bool = True,
        callback: Optional[Callable] = None,
    ) -> Union[Dict, str]:
        """
        Returns results from a completed query, optionally with limit and offset

        :param limit: number of rows to return (default all)
        :param offset: number of rows to skip
        :param sort: gor sort string in format '[column] [ASC|DESC]'
        :param filt: filter to apply to the results serverside
        :param perspective: perspective name to apply to results (for template queries)
        :param is_json: return rows as a dictionary containing 'header' and 'data'
        :returns: dictonary containing 'header' and 'data' lists or tsv
        :raises: QueryError

        """
        start_time = time.time()
        url = self.links["result"]
        if self.status != "DONE":
            raise QueryError(f"Query {self.query_id} is {self.status}")
        if not self.available:
            raise QueryError(
                f"Query results for query {self.query_id} are not available"
            )
        accept = "application/json+compact" if is_json else "text/tab-separated-values"

        responses = []
        num_rows_total = self.line_count or 0
        num_rows_to_fetch = num_rows_total
        if limit:
            num_rows_to_fetch = min(num_rows_total, limit)
        num_rows_received = 0
        skip_header = False
        if num_rows_to_fetch > RESULTS_PAGE_SIZE:
            log.info(
                "Requesting %s rows in %s rows per page...",
                num_rows_to_fetch,
                min(RESULTS_PAGE_SIZE, num_rows_to_fetch),
            )
        while num_rows_received < num_rows_to_fetch:
            num_rows_remaining = num_rows_to_fetch - num_rows_received
            num_rows_to_fetch_this_time = min(num_rows_remaining, RESULTS_PAGE_SIZE)
            data = {
                "limit": num_rows_to_fetch_this_time,
                "offset": num_rows_received,
                "sort": sort,
                "skipheader": skip_header,
                "filt": filt,
                "perspective": perspective,
            }
            skip_header = True
            st = time.time()
            resp = self.session.get(url, json=data, headers={"Accept": accept})
            diff = time.time() - st
            num_rows_received += num_rows_to_fetch_this_time
            responses.append(resp)
            log.debug(
                "Fetched %s rows this time in %.1fsec. Received %s/%s rows",
                num_rows_to_fetch_this_time,
                diff,
                num_rows_received,
                num_rows_to_fetch,
            )
            if callback:
                callback(received=num_rows_received, total=num_rows_total)
        ret: Union[Dict, str]
        if is_json:
            ret = {}
            for r in responses:
                contents = r.json()
                if "data" not in ret:
                    ret["header"] = contents["header"]
                    ret["data"] = []
                ret["data"].extend(contents["data"])
        else:
            ret = ""
            for r in responses:
                ret += r.text
        log.info(
            "Retrieved %s rows from server in %.2f sec",
            num_rows_received,
            (time.time() - start_time),
        )
        return ret

    def cancel(self):
        """
        Cancel a running query

        :raises QueryError: Query is not in a running state
        """
        if self.status not in RUNNING_STATUSES:
            raise QueryError("Query is not running")
        self.session.delete(self.url)

    def dataframe(self, limit: Optional[int] = None):
        """
        Return a Pandas dataframe object containing the results of this query

        :param limit: Maximum number of rows to return (default all)
        :raises QueryError: If the pandas library is not installed
        :return: Pandas dataframe object
        """
        if not jupyter_available():
            raise QueryError("Pandas library is not installed")
        tsv_data = self.get_results(is_json=False, limit=limit)
        import pandas as pd

        if not tsv_data:
            return pd.DataFrame()
        df = pd.read_csv(StringIO(tsv_data), delimiter="\t")  # type: ignore
        return df



#!/usr/bin/env python
import os
import sys

from setuptools import setup, find_packages

if sys.version_info < (3, 6):
    raise NotImplementedError(
        """Genuity Science SDK does not support Python versions older than 3.6"""
    )
version_file = os.path.abspath(os.path.join("nextcode", "VERSION"))

def get_version():
    with open(version_file) as f:
        return f.readlines()[0].strip()

with open("README.md", "r") as fh:
    long_description = fh.read()

version = get_version()

setup(
    name="nextcode-sdk",
    python_requires=">=3.6",
    version=version,
    license="MIT",
    author="WUXI NextCODE",
    author_email="support@wuxinextcode.com",
    description="Python SDK for Genuity Science Services",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.github.com/wuxi-nextcode/nextcode-python-sdk",
    packages=find_packages(exclude=["tests"]),
    package_data={"nextcode": ["VERSION"]},
    install_requires=[
        "python-dateutil",
        "PyYAML",
        "requests",
        "hjson",
        "PyJWT~=1.7.1",
        "boto3",
        "pytest",
    ],
    extras_require={
        "jupyter": ["pandas", "ipython", "termcolor", "tqdm", "ipywidgets", "plotly"]
    },
    include_package_data=True,
    entry_points={},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Jupyter",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    test_suite="tests",
)



"""
Pipeline Job
------------------

The PipelineJob class represents a pipelines job model from the RESTFul Workflow Service API

"""

import json
import datetime
import dateutil
import time
import logging
from typing import Callable, Union, Optional, Dict, List

from . import RUNNING_STATUSES, FINISHED_STATUSES, FAILED_STATUSES
from .exceptions import JobError
from ...exceptions import ServerError
from ...session import ServiceSession

log = logging.getLogger(__name__)


def _smart_cast(name, val):
    if not val:
        return val
    if name.startswith("date_"):
        try:
            val = dateutil.parser.parse(val)
        except Exception:
            pass
    else:
        try:
            val = int(val)
        except:
            try:
                val = float(val)
            except:
                pass
    return val


class PipelineJob:
    """
    Proxy object for a serverside pipeline job.

    This object can be queried for the current status of a pipeline job and will
    automatically refresh from server until the job is finished (or failed)
    """

    def __init__(self, session: ServiceSession, job_id: int, job: Dict):
        self.session = session
        self.job = job
        self.job_id = self.job["job_id"]
        self.links = self.job["links"]

    def __repr__(self) -> str:
        return f"<Pipeline Job {self.job_id} ({self.status})>"

    @property
    def duration(self) -> str:
        """
        Stringified duration of job for human consumption
        """
        if self.complete_date:
            complete_date = self.complete_date
        elif self.status in RUNNING_STATUSES:
            complete_date = datetime.datetime.utcnow()
        elif self.status_date:
            complete_date = self.status_date
        else:
            return "-"
        ret = complete_date - self.submit_date
        # remove microseconds since no one wants them
        ret = ret - datetime.timedelta(microseconds=ret.microseconds)
        return ret

    @property
    def finished(self) -> bool:
        """
        Has the job finished (might be failed)
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in FINISHED_STATUSES

    @property
    def running(self) -> bool:
        """
        Is the job currently running
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in RUNNING_STATUSES

    @property
    def failed(self) -> bool:
        """
        Is the job in a failed state
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status in FAILED_STATUSES

    @property
    def done(self) -> bool:
        """
        Has the job completed successfully
        """
        if self.status in RUNNING_STATUSES:
            self.refresh()
        return self.status == "COMPLETED"

    def refresh(self) -> None:
        """
        Refresh the local cache of the serverside job object
        """
        self.job = self.session.get(self.links["self"]).json()

    def cancel(
        self, status: Optional[str] = None, status_message: Optional[str] = None
    ) -> Optional[str]:
        """
        Cancel a running job

        :param status: status of the cancelled job. Defaults to CANCELLED
        :returns: Status message from the server
        """
        data: Dict = {}
        data["status"] = status or "CANCELLED"
        data["status_message"] = status_message
        resp = self.session.delete(self.links["self"], json=data)
        status_message = resp.json()["status_message"]
        return status_message

    def steps(self) -> List[Dict]:
        """
        Get a list of all nextflow processes in this job
        """
        url = self.links["steps"]
        resp = self.session.get(url)
        steps = resp.json()
        for step in steps:
            for k, v in step.items():
                step[k] = _smart_cast(k, v)
        return steps

    def instance(self) -> Dict:
        """
        Get instance information for this job
        """
        url = self.links["instances"]
        resp = self.session.get(url)
        url = resp.json()[0]["links"]["self"]
        ret = self.session.get(url).json()
        return ret

    def events(self, limit: int = 50) -> List[Dict]:
        """
        Get a list of events reported by Nextflow for this job

        :param limit: Maximum number of events to return
        """
        url = self.links["events"]
        data = {"limit": limit}
        resp = self.session.get(url, json=data)
        return resp.json()["events"]

    def log_groups(self) -> Dict:
        """
        Get available log groups

        :returns: Dictionary with log group name and server url to the log group
        """
        logs_url = self.links["logs"]
        resp = self.session.get(logs_url)
        return resp.json()["links"]

    def logs(self, log_group: str = "pod", log_filter: Optional[str] = None) -> str:
        """
        Get text logs for the specified log group

        :param log_group: Name of the log group to view
        :param log_filter: Optional filter to apply to the logs
        :raises: :exc:`ServerError` If the log group is not available
        """
        groups = self.log_groups()
        url = None
        for k, v in groups.items():
            if k.startswith(log_group):
                url = v
                break
        if not url:
            raise ServerError(f"Log Group '{log_group}' is not available.")
        if log_filter:
            url += "?filter=%s" % log_filter
        logs = self.session.get(url).text
        return logs

    def __getattr__(self, name):
        try:
            val = self.job[name]
        except KeyError:
            raise AttributeError

        val = _smart_cast(name, val)
        return val



import os
import botocore.session
import configparser
from collections import OrderedDict


def find_aws_credentials(profile):
    """
    Returns the aws credentials for the specified profile.
    If no profile is passed in, returns the credentials for the currently selected profile

    Args:
        profile name

    Returns:
        Dict containing at least aws_access_key_id, aws_secret_access_key

    Raises:
        RuntimeError is no default profile or the named profile was not found

    """
    if not profile:
        access_key = None
        secret_key = None
        region = None
        token = ""
        credentials = botocore.session.get_session().get_credentials()
        if credentials:
            access_key = credentials.access_key
            secret_key = credentials.secret_key
            region = credentials.region
            token = getattr(credentials, "token") or ""
        if not access_key or not secret_key:
            raise RuntimeError("No Default AWS profile set")

        ret = {
            "aws_access_key_id": access_key,
            "aws_secret_access_key": secret_key,
            "aws_session_token": token,
        }
        # only add the region if it is defined
        if region:
            ret["region"] = region

        return ret
    else:

        folder = os.path.join(os.path.expanduser("~"), ".aws")
        filename = os.path.join(folder, "credentials")
        cfg = configparser.ConfigParser()
        with open(filename) as fp:
            cfg.read_file(fp)
            ret = {}
            if profile not in cfg:
                raise RuntimeError(
                    "No AWS profile '%s' found in %s" % (profile, filename)
                )
            for key in cfg[profile]:
                ret[key] = cfg[profile][key]
        return ret


def creds_to_dict(credentials: list) -> OrderedDict:
    """
    :param credentials: list of the form ['upload=joe', 'remote_profile=local_profile']
    :return OrderedDict of the form {'upload': 'joe', 'remote_profile': 'local_profile'}
    """
    ret = OrderedDict()
    # credentials are supplied on the form 'upload=joe'
    # 'joe' being the local profile in ~/.aws/credentials
    # and 'upload' being the name we give the credentials
    # when forwarding to the workflow-service
    for cred in credentials:
        key, value = cred.split("=")
        ret[key] = value
    return ret


def generate_credential_struct(credential_map: OrderedDict) -> dict:
    """

    :param credential_map: An OrderedDict of profile maps of the form local-profile-name=remote-profile-name
    :type credential_map: OrderedDict
    :return: Returns an empty dict if credential_map is Falsy, otherwise returns  AWS Credentials as a dict, keyed to profile_name.
    :rtype: dict
    """
    cred_struct = {}
    if not credential_map:
        return cred_struct

    for upload_name, local_name in credential_map.items():
        cred = find_aws_credentials(local_name)
        cred_struct[upload_name] = {}
        cred_struct[upload_name]["aws_access_key_id"] = cred["aws_access_key_id"]
        cred_struct[upload_name]["aws_secret_access_key"] = cred[
            "aws_secret_access_key"
        ]
        if "region" in cred:
            cred_struct[upload_name]["region"] = cred["region"]
    return cred_struct



"""
Workflow Service
------------------
Service object for interfacing with the Workflow service API

"""
INIT_STATUSES = ("PENDING",)
RUNNING_STATUSES = ("PENDING", "STARTED")
FAILED_STATUSES = ("ERROR", "KILLED")
FINISHED_STATUSES = ("CANCELLED", "COMPLETED") + FAILED_STATUSES

from .service import Service
